from typing import Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.grating_coupler_elliptical import grating_tooth_points
from gdsfactory.geometry.functions import DEG2RAD
from gdsfactory.types import Layer, Optional


@gf.cell
def grating_coupler_elliptical_trenches(
    polarization: str = "te",
    fiber_marker_width: float = 11.0,
    fiber_marker_layer: Optional[Layer] = gf.LAYER.TE,
    taper_length: float = 16.6,
    taper_angle: float = 30.0,
    trenches_extra_angle: float = 9.0,
    wavelength: float = 1.53,
    fiber_angle: float = 15.0,
    grating_line_width: float = 0.343,
    wg_width: float = 0.5,
    neff: float = 2.638,  # tooth effective index
    ncladding: float = 1.443,  # cladding index
    layer: Tuple[int, int] = gf.LAYER.WG,
    layer_trench: Tuple[int, int] = gf.LAYER.SLAB150,
    p_start: int = 26,
    n_periods: int = 30,
    end_straight_length: float = 0.2,
) -> Component:
    r"""Returns Grating coupler with defined trenches.

    Some foundries define the grating coupler by a shallow etch step (trenches)
    Others define the slab that they keep (see grating_coupler_elliptical)

    Args:
        polarization: 'te' or 'tm'
        taper_length: taper length from straight I/O
        taper_angle: grating flare angle
        wavelength: grating transmission central wavelength
        fiber_angle: fibre polish angle in degrees
        grating_line_width: of the 220 ridge
        wg_width: straight width
        neff: 2.638  # tooth effective index
        ncladding: cladding index
        layer: for the grating teeth
        layer_trench: for the trench
        p_start:  first tooth
        n_periods: number of periods
        end_straight_length: at the end of


    .. code::

                      fiber

                   /  /  /  /
                  /  /  /  /
                _|-|_|-|_|-|___
        WG  o1  ______________|

    """

    # Define some constants

    # Compute some ellipse parameters
    sthc = np.sin(fiber_angle * DEG2RAD)
    d = neff ** 2 - ncladding ** 2 * sthc ** 2
    a1 = wavelength * neff / d
    b1 = wavelength / np.sqrt(d)
    x1 = wavelength * ncladding * sthc / d

    a1 = round(a1, 3)
    b1 = round(b1, 3)
    x1 = round(x1, 3)

    period = float(a1 + x1)
    trench_line_width = period - grating_line_width

    c = gf.Component()

    # Make each grating line
    for p in range(p_start, p_start + n_periods + 1):
        pts = grating_tooth_points(
            p * a1,
            p * b1,
            p * x1,
            width=trench_line_width,
            taper_angle=taper_angle + trenches_extra_angle,
        )
        c.add_polygon(pts, layer_trench)

    # Make the taper
    p_taper = p_start - 1
    p_taper_eff = p_taper
    a_taper = a1 * p_taper_eff
    # b_taper = b1 * p_taper_eff
    x_taper = x1 * p_taper_eff
    x_output = a_taper + x_taper - taper_length + grating_line_width / 2

    xmax = x_output + taper_length + n_periods * period + 3
    y = wg_width / 2 + np.tan(taper_angle / 2 * np.pi / 180) * xmax
    pts = [
        (x_output, -wg_width / 2),
        (x_output, wg_width / 2),
        (xmax, y),
        (xmax + end_straight_length, y),
        (xmax + end_straight_length, -y),
        (xmax, -y),
    ]
    c.add_polygon(pts, layer)

    x = np.round(taper_length + period * n_periods / 2, 3)
    if fiber_marker_layer:
        circle = gf.components.circle(
            radius=fiber_marker_width / 2, layer=fiber_marker_layer
        )
        circle_ref = c.add_ref(circle)
        circle_ref.movex(x)
    name = f"vertical_{polarization.lower()}"
    c.add_port(
        name=name,
        midpoint=[x, 0],
        width=fiber_marker_width,
        orientation=0,
        layer=fiber_marker_layer,
        port_type=name,
    )

    c.add_port(
        name="o1", midpoint=[x_output, 0], width=wg_width, orientation=180, layer=layer
    )
    c.info.period = period
    c.info.polarization = polarization
    c.info.wavelength = wavelength
    return c


grating_coupler_te = gf.partial(
    grating_coupler_elliptical_trenches, polarization="te", taper_angle=35
)

grating_coupler_tm = gf.partial(
    grating_coupler_elliptical_trenches,
    polarization="tm",
    neff=1.8,
    grating_line_width=0.6,
    fiber_marker_layer=gf.LAYER.TM,
)


if __name__ == "__main__":
    # c = grating_coupler_elliptical_trenches(polarization="TE")
    # print(c.polarization)
    # c = grating_coupler_te(end_straight_length=10, fiber_marker_layer=None)
    c = grating_coupler_tm()
    print(c.ports.keys())
    c.show()



from gdsfactory.simulation.modes import coupler, waveguide
from gdsfactory.simulation.modes.find_coupling_vs_gap import (
    find_coupling_vs_gap,
    plot_coupling_vs_gap,
)
from gdsfactory.simulation.modes.find_mode_dispersion import find_mode_dispersion
from gdsfactory.simulation.modes.find_modes import find_modes
from gdsfactory.simulation.modes.find_neff_vs_width import (
    find_neff_vs_width,
    plot_neff_vs_width,
)
from gdsfactory.simulation.modes.get_mode_solver_coupler import get_mode_solver_coupler
from gdsfactory.simulation.modes.get_mode_solver_rib import get_mode_solver_rib

__all__ = [
    "find_modes",
    "find_neff_vs_width",
    "find_mode_dispersion",
    "find_coupling_vs_gap",
    "get_mode_solver_rib",
    "get_mode_solver_coupler",
    "plot_neff_vs_width",
    "plot_coupling_vs_gap",
    "coupler",
    "waveguide",
]
__version__ = "0.0.2"



""" Command line interface for gdsfactory."""
import os
import pathlib
import shlex
import subprocess
from typing import Optional

import click
from click.core import Context, Option

import gdsfactory
import gdsfactory.build as pb
from gdsfactory.config import CONFIG, print_config
from gdsfactory.gdsdiff.gdsdiff import gdsdiff
from gdsfactory.import_gds import write_cells as write_cells_to_separate_gds
from gdsfactory.install import install_gdsdiff, install_generic_tech, install_klive
from gdsfactory.layers import lyp_to_dataclass
from gdsfactory.mask.merge_json import merge_json
from gdsfactory.mask.merge_markdown import merge_markdown
from gdsfactory.mask.merge_test_metadata import merge_test_metadata
from gdsfactory.mask.write_labels import write_labels

# from gdsfactory.write_doe_from_yaml import write_doe_from_yaml
from gdsfactory.sweep.write_sweep_from_yaml import import_custom_doe_factories
from gdsfactory.tech import LAYER
from gdsfactory.types import PathType

VERSION = "3.8.8"
log_directory = CONFIG.get("log_directory")
cwd = pathlib.Path.cwd()
LAYER_LABEL = LAYER.LABEL


def print_version(ctx: Context, param: Option, value: bool) -> None:
    """Prints the version"""
    if not value or ctx.resilient_parsing:
        return
    click.echo(VERSION)
    ctx.exit()


@click.command(name="delete")
@click.argument("logfile", default="main", required=False)
def log_delete(logfile: str) -> None:
    """Deletes logs"""
    if not os.path.exists(log_directory):
        print("No logs found.")
        return

    filename = os.path.join(log_directory, "{}.log".format(logfile))
    subprocess.check_output(["rm", filename])


# TOOL


@click.group()
def tool() -> None:
    """Commands working with gdsfactory tool"""
    pass


@click.command(name="config")
@click.argument("key", required=False, default=None)
def config_get(key: str) -> None:
    """Shows key values from CONFIG"""
    print_config(key)


# GDS


@click.group()
def gds() -> None:
    """Commands for dealing with GDS files."""
    pass


@click.command(name="layermap_to_dataclass")
@click.argument("filepath", type=click.Path(exists=True))
@click.option("--force", "-f", default=False, help="Force deletion", is_flag=True)
def layermap_to_dataclass(filepath, force: bool) -> None:
    """Converts klayout LYP to a dataclass"""
    filepath_lyp = pathlib.Path(filepath)
    filepath_py = filepath_lyp.with_suffix(".py")
    if not filepath_lyp.exists():
        raise FileNotFoundError(f"{filepath_lyp} not found")
    if not force and filepath_py.exists():
        raise FileExistsError(f"found {filepath_py}")
    lyp_to_dataclass(lyp_filepath=filepath_lyp)


@click.command(name="write_cells")
@click.argument("gdspath")
def write_cells(gdspath) -> None:
    """Write each all level cells into separate GDS files."""
    write_cells_to_separate_gds(gdspath)


@click.command(name="merge_gds_from_directory")
@click.argument("dirpath", required=False, default=None)
@click.argument("gdspath", required=False, default=None)
def merge_gds_from_directory(
    dirpath: Optional[PathType] = None, gdspath: Optional[PathType] = None
) -> None:
    """Merges GDS cells from a directory into a single GDS."""
    dirpath = dirpath or pathlib.Path.cwd()
    gdspath = gdspath or pathlib.Path.cwd() / "merged.gds"
    c = gdsfactory.read.gdsdir(dirpath=dirpath)
    c.write_gds(gdspath=gdspath)
    c.show()


# MASKS


@click.group()
def mask() -> None:
    """Commands for building masks"""
    pass


@click.command(name="clean")
@click.option("--force", "-f", default=False, help="Force deletion", is_flag=True)
def build_clean(force: bool) -> None:
    """Deletes the build folder and contents"""
    message = "Delete {}. Are you sure?".format(CONFIG["build_directory"])
    if force or click.confirm(message, default=True):
        pb.build_clean()


@click.command(name="build_devices")
@click.argument("regex", required=False, default=".*")
def build_devices(regex) -> None:
    """Build all devices described in devices/"""
    pb.build_devices(regex)


@click.command(name="build_does")
@click.argument("yamlpath")
def build_does(yamlpath: str) -> None:
    """Build does defined in sweep.yml"""
    print("this is deprecated")
    import_custom_doe_factories()
    # write_doe_from_yaml()
    pb.build_does(yamlpath)


@click.command(name="write_metadata")
@click.argument("label_layer", required=False, default=LAYER_LABEL)
def mask_merge(label_layer) -> None:
    """merge JSON/Markdown from build/devices into build/mask"""

    gdspath = CONFIG["mask_gds"]
    write_labels(gdspath=gdspath, label_layer=label_layer)

    merge_json()
    merge_markdown()
    merge_test_metadata(gdspath=gdspath)


@click.command(name="write_labels")
@click.argument("gdspath", default=None)
@click.argument("label_layer", required=False, default=LAYER_LABEL)
def write_mask_labels(gdspath: str, label_layer) -> None:
    """Find test and measurement labels."""
    if gdspath is None:
        gdspath = CONFIG["mask_gds"]

    write_labels(gdspath=gdspath, label_layer=label_layer)


# EXTRA


@click.command()
@click.argument("filename")
def show(filename: str) -> None:
    """Show a GDS file using klive"""
    gdsfactory.show(filename)


@click.command()
@click.argument("gdspath1")
@click.argument("gdspath2")
@click.option("--xor", "-x", default=False, help="include xor", is_flag=True)
def diff(gdspath1: str, gdspath2: str, xor: bool = False) -> None:
    """Show boolean difference between two GDS files."""
    diff = gdsdiff(str(gdspath1), str(gdspath2), xor=xor)
    diff.show()


@click.command()
def install() -> None:
    """Install Klive and generic tech layermap."""
    install_generic_tech()
    install_klive()
    install_gdsdiff()


@click.command()
def test() -> None:
    """Run tests using pytest.
    You can also just run `pytest` directly."""

    os.chdir(CONFIG["repo_path"])
    command = shlex.split("pytest")
    subprocess.call(command)


@click.group()
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Show the version number.",
)
def gf():
    """`gf` is the command line tool for gdsfactory.
    It helps you work with GDS files.
    """
    pass


mask.add_command(build_clean)
mask.add_command(build_devices)
mask.add_command(build_does)
mask.add_command(mask_merge)
mask.add_command(write_mask_labels)

gds.add_command(layermap_to_dataclass)
gds.add_command(write_cells)
gds.add_command(merge_gds_from_directory)
gds.add_command(show)
gds.add_command(diff)

tool.add_command(config_get)
tool.add_command(test)
tool.add_command(install)

gf.add_command(gds)
gf.add_command(tool)
gf.add_command(mask)


if __name__ == "__main__":
    gf()



from typing import Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.tech import LAYER


@gf.cell
def hline(
    length: float = 10.0,
    width: float = 0.5,
    layer: Tuple[int, int] = LAYER.WG,
    port_type: str = "optical",
) -> Component:
    """Horizonal line straight, with ports on east and west sides"""
    c = gf.Component()
    a = width / 2
    if length > 0 and width > 0:
        c.add_polygon([(0, -a), (length, -a), (length, a), (0, a)], layer=layer)

    c.add_port(
        name="o1",
        midpoint=[0, 0],
        width=width,
        orientation=180,
        layer=layer,
        port_type=port_type,
    )
    c.add_port(
        name="o2",
        midpoint=[length, 0],
        width=width,
        orientation=0,
        layer=layer,
        port_type=port_type,
    )

    c.info.width = float(width)
    c.info.length = float(length)
    return c


if __name__ == "__main__":
    c = hline(width=10)
    print(c)
    c.show()



import pathlib
from pathlib import Path
from typing import Dict, Tuple

from gdsfactory.component import Component
from gdsfactory.config import CONFIG
from gdsfactory.name import dict2name, get_name_short
from gdsfactory.tech import LAYER


def get_sparameters_path(
    component: Component,
    layer_to_material: Dict[Tuple[int, int], str],
    layer_to_thickness: Dict[Tuple[int, int], int],
    dirpath: Path = CONFIG["sparameters"],
    **kwargs,
) -> Path:
    """Returns Sparameters filepath.

    it only includes the layers that are present in the component

    Args:
        component:
        layer_to_material: GDSlayer tuple to material alias
        layer_to_thickness: GDSlayer tuple to thickness (nm)
        dirpath:
        kwargs: simulation settings
    """
    dirpath = pathlib.Path(dirpath)
    dirpath = (
        dirpath / component.function_name
        if hasattr(component, "function_name")
        else dirpath
    )
    dirpath.mkdir(exist_ok=True, parents=True)
    material_to_thickness = {
        layer_to_material[layer]: layer_to_thickness[layer]
        for layer in layer_to_thickness.keys()
        if tuple(layer) in component.get_layers()
    }
    material_to_thickness.update(**kwargs)
    suffix = get_name_short(dict2name(**material_to_thickness))
    return dirpath / f"{component.name}_{suffix}.dat"


def test_get_sparameters_path() -> None:
    import gdsfactory as gf

    layer_to_thickness_sample = {
        LAYER.WG: 220e-3,
        LAYER.SLAB90: 90e-3,
    }
    layer_to_material_sample = {
        LAYER.WG: "si",
        LAYER.SLAB90: "si",
    }

    c = gf.components.straight()
    p = get_sparameters_path(
        component=c,
        layer_to_thickness=layer_to_thickness_sample,
        layer_to_material=layer_to_material_sample,
    )
    assert p.stem == f"{c.name}_si220n", p.stem

    c = gf.components.straight(layer=LAYER.SLAB90)
    p = get_sparameters_path(
        c,
        layer_to_thickness=layer_to_thickness_sample,
        layer_to_material=layer_to_material_sample,
    )
    cell_name = f"{c.name}_si90n"
    assert p.stem == cell_name, p.stem


if __name__ == "__main__":
    # import gdsfactory as gf
    # c = gf.components.straight()
    # p = get_sparameters_path(c)
    # print(p)

    test_get_sparameters_path()



import functools
from pathlib import Path
from typing import Union

import klayout.db as pya
from klayout.dbcore import Cell, Layout

CELLS = {}


@functools.lru_cache()
def load_gds(filepath: Union[Path, str]) -> Layout:
    filepath = str(filepath)
    layout = pya.Layout()
    try:
        layout.read(filepath)
    except RuntimeError as e:
        print(f"Error reading {filepath}")
        raise e
    cell = layout.top_cell()
    cell.metadata = {}

    # To make sure the cell does not get destroyed
    global CELLS
    CELLS[cell.name] = layout
    return layout


def import_cell(layout: Layout, cell: Cell) -> Cell:
    """Imports a cell from another Layout into a given layout"""
    # If the cell is already in the library, skip loading
    if layout.cell(cell.name):
        return layout.cell(cell.name)

    # Create a holder cell and copy in the shapes
    new_cell = layout.create_cell(cell.name)
    new_cell.copy_shapes(cell)

    # Import all the child cells
    for child_index in cell.each_child_cell():
        import_cell(layout, cell.layout().cell(child_index))

    # Import all of the instances, doing the mapping from Layout to Layout
    for instance in cell.each_inst():
        cell_index = layout.cell(instance.cell.name).cell_index()
        # new_instance = pya.CellInstArray(cell_index, instance.trans)
        new_instance = instance.cell_inst.dup()
        new_instance.cell_index = cell_index
        new_cell.insert(new_instance)
    return new_cell



from functools import partial

import meep as mp
import meep.materials as mat

MATERIALS = [m for m in dir(mat) if not m.startswith("_")]


def get_index(
    wavelength: float = 1.55,
    name: str = "Si",
) -> float:
    """Returns refractive index from Meep's material database.

    Args:
        name: material name
        wavelength: wavelength (um)

    """
    if name not in MATERIALS:
        raise ValueError(f"{name} not in {MATERIALS}")

    medium = getattr(mat, name)
    epsilon_matrix = medium.epsilon(1 / wavelength)
    epsilon11 = epsilon_matrix[0][0]
    return epsilon11 ** 0.5


def get_material(
    wavelength: float = 1.55,
    name: str = "Si",
) -> mp.Medium:
    """Returns Meep Medium from database

    Args:
        name: material name
        wavelength: wavelength (um)

    """

    # FIXME: need to remove this. If I remove this, then I get no fields.
    if name == "SiO2":
        return mp.Medium(epsilon=2.25)
    elif name.lower() == "si":
        return mp.Medium(epsilon=12)
    elif name.lower() == "sin":
        return mp.Medium(epsilon=4)
    else:
        raise ValueError(f"material, name = {name} not in {MATERIALS}")

    return getattr(mat, name)


si = partial(get_index, name="Si")

sio2 = partial(get_index, name="SiO2")


if __name__ == "__main__":
    print(get_index(name="Si"))



from gdsfactory.add_padding import get_padding_points
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bezier import bezier
from gdsfactory.cross_section import strip
from gdsfactory.port import auto_rename_ports
from gdsfactory.types import CrossSectionFactory, Float2


@cell
def bend_s(
    size: Float2 = (10.0, 2.0),
    nb_points: int = 99,
    with_cladding_box: bool = True,
    cross_section: CrossSectionFactory = strip,
    **kwargs
) -> Component:
    """S bend with bezier curve

    stores min_bend_radius property in self.info['min_bend_radius']
    min_bend_radius depends on height and length

    Args:
        size: in x and y direction
        nb_points: number of points
        with_cladding_box: square bounding box to avoid DRC errors
        cross_section: function
        kwargs: cross_section settings

    """
    dx, dy = size
    x = cross_section(**kwargs)
    width = x.info["width"]
    layer = x.info["layer"]

    c = Component()

    bend = bezier(
        width=width,
        control_points=[(0, 0), (dx / 2, 0), (dx / 2, dy), (dx, dy)],
        npoints=nb_points,
        layer=layer,
    )

    bend_ref = c << bend
    c.add_ports(bend_ref.ports)
    c.copy_child_info(bend)
    c.info.start_angle = bend.info.start_angle
    c.info.end_angle = bend.info.end_angle
    c.info.length = bend.info.length
    c.info.min_bend_radius = bend.info.min_bend_radius

    if with_cladding_box and x.info["layers_cladding"]:
        layers_cladding = x.info["layers_cladding"]
        cladding_offset = x.info["cladding_offset"]
        points = get_padding_points(
            component=c,
            default=0,
            bottom=cladding_offset,
            top=cladding_offset,
        )
        for layer in layers_cladding or []:
            c.add_polygon(points, layer=layer)

    auto_rename_ports(c)
    return c


if __name__ == "__main__":
    c = bend_s(width=1, layers_cladding=[(2, 0)])
    c = bend_s(size=[10, 2.5], layers_cladding=[(2, 0)])  # 10um bend radius
    c = bend_s(size=[20, 3], layers_cladding=[(2, 0)])  # 10um bend radius
    c.pprint()
    # c = bend_s_biased()
    # print(c.info["min_bend_radius"])
    c.show()



from pathlib import Path

import pytest
from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.import_gds import add_ports_from_markers_center, import_gds
from gdsfactory.port import auto_rename_ports

# gdspaths = [gf.CONFIG["gdsdir"] / name for name in ["mmi1x2.gds", "mzi2x2.gds"]]
gdspaths = [gf.CONFIG["gdsdir"] / name for name in ["mzi2x2.gds"]]


@pytest.mark.parametrize("gdspath", gdspaths)
def test_components_ports(
    gdspath: Path, data_regression: DataRegressionFixture
) -> gf.Component:
    c = import_gds(gdspath)
    add_ports_from_markers_center(c)
    auto_rename_ports(c)
    data_regression.check(c.to_dict())


if __name__ == "__main__":
    c = import_gds(gdspaths[0])
    add_ports_from_markers_center(c)
    auto_rename_ports(c)
    print(c.ports.keys())
    print(c.name)



from typing import Tuple

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.straight import straight


@cell
def array(
    component: gf.types.ComponentOrFactory = straight,
    spacing: Tuple[float, float] = (150.0, 150.0),
    columns: int = 6,
    rows: int = 1,
) -> Component:
    """Returns an array of components.

    Args:
        component: to replicate
        spacing: x, y spacing
        columns: in x
        rows: in y

    Raises:
        ValueError: If columns > 1 and spacing[0] = 0
        ValueError: If rows > 1 and spacing[1] = 0

    .. code::

        2 rows x 4 columns
         ___        ___       ___          ___
        |   |      |   |     |   |        |   |
        |___|      |___|     |___|        |___|


         ___        ___       ___          ___
        |   |      |   |     |   |        |   |
        |___|      |___|     |___|        |___|


    """
    if rows > 1 and spacing[1] == 0:
        raise ValueError(f"rows = {rows} > 1 require spacing[1] > 0")

    if columns > 1 and spacing[0] == 0:
        raise ValueError(f"columns = {columns} > 1 require spacing[0] > 0")

    c = Component()
    component = component() if callable(component) else component
    c.add_array(component, columns=columns, rows=rows, spacing=spacing)

    for col in range(columns):
        for row in range(rows):
            for port in component.ports.values():
                name = f"{port.name}_{row+1}_{col+1}"
                c.add_port(name, port=port)
                c.ports[name].move((col * spacing[0], row * spacing[1]))
    return c


if __name__ == "__main__":

    c2 = array(rows=2, columns=2, spacing=(100, 100))
    c2.show(show_ports=True)



import numpy as np

from gdsfactory.simulation.modes.find_mode_dispersion import find_mode_dispersion
from gdsfactory.simulation.modes.get_mode_solver_rib import get_mode_solver_rib


def test_find_modes_dispersion():
    ms = get_mode_solver_rib(wg_width=0.45)
    modes = find_mode_dispersion(mode_solver=ms)
    m1 = modes

    # print(m1.neff)
    # print(m1.ng)

    neff1 = 2.3294606863357443
    ng1 = 4.112495445787479

    assert np.isclose(m1.neff, neff1), (m1.neff, neff1)
    assert np.isclose(m1.ng, ng1), (m1.ng, ng1)


if __name__ == "__main__":
    test_find_modes_dispersion()



"""Update GDS view in Klayout dynamically.
Requires the Klayout plugin installed in Klayout.
This happens when you run `pf install`.
"""

import json
import os
import socket
from pathlib import Path
from typing import Union


def show(gds_filename: Union[Path, str], keep_position: bool = True) -> None:
    """Show GDS in klayout."""
    if not os.path.isfile(gds_filename):
        raise ValueError(f"{gds_filename} does not exist")
    data = {
        "gds": os.path.abspath(gds_filename),
        "keep_position": keep_position,
    }
    data_string = json.dumps(data)
    try:
        conn = socket.create_connection(("127.0.0.1", 8082), timeout=1.0)
        data_string = data_string + "\n"
        data_string = (
            data_string.encode() if hasattr(data_string, "encode") else data_string
        )
        conn.sendall(data_string)
        conn.close()
    except socket.error:
        pass


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.straight()
    gdspath = c.write_gds()
    show(gdspath)



# type: ignore
""" gdspy based placer

YAML defines component DOE settings and placement

.. code:: yaml

    placer:
        width: 30000  # um
        height: 20000 # um
        x_spacing: 1000 # x spacing origin to origin between the components
        y_spacing: 500 #  y spacing origin to origin between the components
        x_start: 500
        y_start: 100

    mmi1x2_gap:
        doe_name: doe1
        component: mmi1x2
        gap: [0.5, 0.6]
        length: 10

    mmi1x2:
        doe_name: doe1
        component: mmi1x2
        length: [11, 12]
        gap: [0.2, 0.3]
        do_permutation: False

    placement:
        A-B1-2: doe1
"""

import os
import pathlib
import sys
from pathlib import Path
from typing import Callable, Dict, List, Union

from omegaconf import OmegaConf

import gdsfactory as gf
from gdsfactory.component import Component, ComponentReference
from gdsfactory.components import factory
from gdsfactory.config import CONFIG
from gdsfactory.sweep.read_sweep import get_settings_list, read_sweep
from gdsfactory.types import ComponentFactoryDict


def _print(*args, **kwargs):
    print(*args, **kwargs)
    sys.stdout.flush()


def placer_grid_cell_refs(
    component_factory: ComponentFactoryDict = factory,
    cols: int = 1,
    rows: int = 1,
    dx: float = 10.0,
    dy: float = 10.0,
    x0: float = 0,
    y0: float = 0,
    **settings,
) -> List[ComponentReference]:
    if callable(component_factory):
        settings_list = get_settings_list(**settings)
        component_list = [component_factory(**s) for s in settings_list]
    else:
        component_list = component_factory

    indices = [(i, j) for j in range(cols) for i in range(rows)]

    if rows * cols < len(component_list):
        raise ValueError(
            "Shape ({}, {}): Not enough emplacements ({}) for all these components ({}).".format(
                rows, cols, len(indices), len(component_list)
            )
        )
    components = []
    for component, (i, j) in zip(component_list, indices):
        c_ref = component.ref(position=(x0 + j * dx, y0 + i * dy))
        components += [c_ref]

    return components


def pack_horizontal(
    cells,
    row_ids=None,
    x0: float = 0.0,
    y0: float = 0.0,
    align_x: str = "W",
    align_y: str = "S",
    margin_x: float = 20.0,
    margin_y: float = 20.0,
):
    """
    Args:
        cells: a list of cells  (size n)
        row_ids: a list of row ids (size n)
            where each id represents the row where the cell should be placed
            None by default => all cells in the same row

    returns a list of cell references
    """
    heights = [c.size_info.height for c in cells]

    row_ids = row_ids or [0] * len(cells)

    if len(cells) != len(row_ids):
        raise ValueError(
            "Each cell should be assigned a row id. \
        Got {} cells for {} row ids".format(
                len(cells), len(row_ids)
            )
        )

    # Find the height of each row to fit the cells
    # Also group the cells by row

    unique_row_ids = list(set(row_ids))
    unique_row_ids.sort()
    _row_to_heights = {r: [] for r in set(row_ids)}
    row_to_cells = {r: [] for r in unique_row_ids}
    for row, h, cell in zip(row_ids, heights, cells):
        _row_to_heights[row] += [h]
        row_to_cells[row] += [cell]

    row_to_height = {k: max(v) for k, v in _row_to_heights.items()}

    components = []

    # Do the packing per row
    y = y0
    for row in unique_row_ids:
        cells = row_to_cells[row]
        x = x0

        for c in cells:
            if align_x == "W" and align_y == "S":
                component_origin = c.size_info.sw
            elif align_x == "E" and align_y == "S":
                component_origin = c.size_info.se
            elif align_x == "E" and align_y == "N":
                component_origin = c.size_info.ne
            elif align_x == "W" and align_y == "N":
                component_origin = c.size_info.nw
            try:
                components += [c.ref(position=-component_origin + (x, y))]
            except ValueError as e:
                if align_x not in ["W", "E"]:
                    print("align_x should be `W`, `E` or a float")
                if align_y not in ["N", "S"]:
                    print("align_y should be `N`, `S` or a float")
                raise e

            if align_x == "W":
                x += c.size_info.width + margin_x
            else:
                x += -c.size_info.width - margin_x

        if align_y == "S":
            y += row_to_height[row] + margin_y
        else:
            y += -row_to_height[row] - margin_y

    return components


def pack_vertical(
    cells,
    col_ids=None,
    x0: float = 0.0,
    y0: float = 0.0,
    align_x: str = "W",
    align_y: str = "S",
    margin_x: float = 20.0,
    margin_y: float = 20.0,
) -> List[Component]:
    """
    Args:
        cells: a list of cells  (size n)
        col_ids: a list of column ids (size n)
            where each id represents the row where the cell should be placed
            None by default => all cells are packed in the same column

    returns a list of cell references
    """
    widths = [c.size_info.width for c in cells]
    col_ids = col_ids or [0] * len(cells)

    if len(cells) != len(col_ids):
        raise ValueError(
            "Each cell should be assigned a row id. \
        Got {} cells for {} col ids".format(
                len(cells), len(col_ids)
            )
        )

    # Find the width of each column to fit the cells
    # Also group the cells by column

    unique_col_ids = list(set(col_ids))
    unique_col_ids.sort()
    _col_to_widths = {r: [] for r in set(col_ids)}
    col_to_cells = {r: [] for r in unique_col_ids}
    for col, w, cell in zip(col_ids, widths, cells):
        _col_to_widths[col] += [w]
        col_to_cells[col] += [cell]

    col_to_width = {k: max(v) for k, v in _col_to_widths.items()}

    components = []

    # Do the packing per column
    x = x0
    for col in unique_col_ids:
        cells = col_to_cells[col]
        y = y0

        for c in cells:
            if align_x == "W" and align_y == "S":
                component_origin = c.size_info.sw
            elif align_x == "E" and align_y == "S":
                component_origin = c.size_info.se
            elif align_x == "E" and align_y == "N":
                component_origin = c.size_info.ne
            elif align_x == "W" and align_y == "N":
                component_origin = c.size_info.nw
            components += [c.ref(position=-component_origin + (x, y))]

            if align_y == "S":
                y += c.size_info.height + margin_y
            else:
                y += -c.size_info.height - margin_y

        if align_x == "W":
            x += col_to_width[col] + margin_x
        else:
            x += -col_to_width[col] - margin_x

    return components


def placer_fixed_coords(cells, x, y, **kwargs):
    if len(cells) == 1:
        cells = cells * len(x)

    return [c.ref(position=(_x, _y)) for c, _x, _y in zip(cells, x, y)]


PLACER_NAME2FUNC = {
    "grid": placer_grid_cell_refs,
    "pack_row": pack_horizontal,
    "pack_col": pack_vertical,
    "fixed_coords": placer_fixed_coords,
}


def load_placer_with_does(filepath, defaults=None):
    """load placer settings

    Args:
        filepath: a yaml file containing the does and placer information

    Returns:
        a dictionnary of DOEs with:
        {
            doe_name1: [(component_factory_name, parameters), ...]
            doe_name2: [(component_factory_name, parameters), ...]
            ...
        }

    """
    defaults = defaults or {"do_permutation": True}
    does = {}
    data = OmegaConf.load(filepath)

    placer_info = data.pop("placer")
    component_placement = data.pop("placement")
    gds_files = {}
    if "gds" in data.keys():
        gds_files.update(data.pop("gds"))

    for doe_chunk in data.values():
        doe_name = doe_chunk.pop("doe_name")
        component_type = doe_chunk.pop("component")

        if doe_name not in does:
            does[doe_name] = {"settings": [], "component_type": component_type}

        do_permutation = defaults["do_permutation"]
        if "do_permutation" in doe_chunk:
            do_permutation = doe_chunk.pop("do_permutation")

        doe = does[doe_name]
        # All the remaining parameters are component parameters
        if "settings" in doe_chunk:
            settings = doe_chunk.pop("settings")
        else:
            settings = {}
        doe["list_settings"] += get_settings_list(do_permutation, **settings)

        # check that the sweep is valid (only one type of component)
        assert (
            component_type == doe["component_type"]
        ), "There can be only one component type per sweep. Got {} while expecting {}".format(
            component_type, doe["component_type"]
        )

    return does, placer_info, component_placement, gds_files


CONTENT_SEP = " , "


def save_doe(
    doe_name, components, doe_root_path=CONFIG["cache_doe_directory"], precision=1e-9
):
    """
    Save all components from this DOE in a tmp cache folder
    """
    doe_dir = pathlib.Path(doe_root_path) / doe_name
    doe_dir.mkdir(parents=True, exist_ok=True)

    # Store list of component names - order matters
    component_names = [c.name for c in components]
    content_file = doe_dir / "content.txt"
    with open(content_file, "w") as fw:
        fw.write(CONTENT_SEP.join(component_names))

    for c in components:
        gdspath = doe_dir / f"{c.name}.gds"
        c.write_gds_with_metadata(gdspath=gdspath, precision=precision)


def load_doe_from_cache(doe_name, doe_root_path=None):
    """
    Load all components for this DOE from the cache
    """
    if doe_root_path is None:
        doe_root_path = CONFIG["cache_doe_directory"]
    doe_dir = os.path.join(doe_root_path, doe_name)
    content_file = os.path.join(doe_dir, "content.txt")
    with open(content_file) as f:
        component_names = f.read().split(CONTENT_SEP)

    gdspaths = [os.path.join(doe_dir, name + ".gds") for name in component_names]
    components = [gf.import_gds(gdspath) for gdspath in gdspaths]
    return components


def load_doe_component_names(doe_name, doe_root_path=None):
    if doe_root_path is None:
        doe_root_path = CONFIG["cache_doe_directory"]
    doe_dir = os.path.join(doe_root_path, doe_name)
    content_file = os.path.join(doe_dir, "content.txt")

    with open(content_file) as f:
        component_names = f.read().split(CONTENT_SEP)
    return component_names


def doe_exists(
    doe_name: str,
    list_settings: Union[List[Dict[str, Union[float, int]]], List[Dict[str, int]]],
    doe_root_path: None = None,
) -> bool:
    """
    Check whether the folder exists and that the number of items in content.txt
    matches the number of items in list_settings
    """
    if doe_root_path is None:
        doe_root_path = CONFIG["cache_doe_directory"]
    doe_dir = os.path.join(doe_root_path, doe_name)
    content_file = os.path.join(doe_dir, "content.txt")
    if not os.path.exists(content_file):
        return False
    with open(content_file) as f:
        component_names = f.read().split(CONTENT_SEP)

    if len(component_names) == len(list_settings) or (
        len(list_settings) == 0 and len(component_names) == 1
    ):
        return True

    print(
        "doe_exists - DOE",
        doe_name,
        "needs regeneration",
        len(component_names),
        len(list_settings),
    )
    return False


def component_grid_from_yaml(filepath: Path, precision: float = 1e-9) -> Component:
    """Returns a Component composed of DOEs/components given in a yaml file
    allows for each DOE to have its own x and y spacing (more flexible than method1)
    """
    input_does = OmegaConf.load(str(filepath))
    mask_settings = input_does["mask"]
    does = read_sweep(filepath)

    placed_doe = None
    placed_does = {}
    if mask_settings.get("name"):
        component_grid = gf.Component(mask_settings["name"])
    else:
        component_grid = gf.Component()

    default_cache_enabled = (
        mask_settings["cache_enabled"] if "cache_enabled" in mask_settings else False
    )
    for doe_name, doe in does.items():
        list_settings = doe["settings"]
        component_type = doe["component"]

        # description = sweep["description"] if "description" in sweep else ''
        # test = sweep["test"] if "test" in sweep else {}
        # analysis = sweep["analysis"] if "analysis" in sweep else {}

        # Get DOE policy concerning the cache
        cache_enabled = (
            doe["cache_enabled"] if "cache_enabled" in doe else default_cache_enabled
        )

        components = None

        # If cache enabled, attempt to load from cache
        if cache_enabled:
            try:
                components = load_doe_from_cache(doe_name)
            except Exception as e:
                _print(e)
                components = None

        # If no component is loaded, build them
        if components is None:
            print("{} - Generating components...".format(doe_name))
            components = build_components(component_type, list_settings)

            # After building the components, if cache enabled, save them
            if cache_enabled:
                save_doe(doe_name, components, precision=precision)
        else:
            _print("{} - Loaded components from cache".format(doe_name))

        # _print(doe_name, [c.name for c in components])
        # Find placer information

        default_settings = {"align_x": "W", "align_y": "S", "margin": 10}

        if "placer" in doe:
            placer_type = doe["placer"].pop("type")
            _placer = PLACER_NAME2FUNC[placer_type]

            # All other attributes are assumed to be settings for the placer
            settings = default_settings.copy()
            settings.update(doe["placer"])

            # x0, y0 can either be float or string
            x0 = settings.pop("x0")
            y0 = settings.pop("y0")

            # Check whether we are doing relative or absolute placement
            if (x0 in ["E", "W"] or y0 in ["N", "S"]) and not placed_doe:
                raise ValueError(
                    "At least one DOE must be placed to use\
                relative placement"
                )

            # For relative placement (to previous DOE)
            if "margin_x" not in settings:
                settings["margin_x"] = settings["margin"]
            if "margin_y" not in settings:
                settings["margin_y"] = settings["margin"]

            margin_x = settings["margin_x"]
            margin_y = settings["margin_y"]
            align_x = settings["align_x"]
            align_y = settings["align_y"]

            # Making sure that the alignment is sensible depending on how we stack

            # If we specify a DOE to place next to, use it
            if "next_to" in settings:
                placed_doe = placed_does[settings.pop("next_to")]

            # Otherwise, use previously placed DOE as starting point
            if x0 == "E":
                x0 = placed_doe.size_info.east
                if align_x == "W":
                    x0 += margin_x

            if x0 == "W":
                x0 = placed_doe.size_info.west
                if align_x == "E":
                    x0 -= margin_x

            if y0 == "N":
                y0 = placed_doe.size_info.north
                if align_y == "S":
                    y0 += margin_y

            if y0 == "S":
                y0 = placed_doe.size_info.south
                if align_y == "N":
                    y0 -= margin_y

            # Add x0, y0 in settings as float
            settings["x0"] = x0
            settings["y0"] = y0

            placed_components = _placer(components, **settings)
        else:
            # If no placer is specified, we assume this is a grid placer
            cols, rows = doe["shape"]
            x0, y0 = doe["origin"]
            dx, dy = doe["spacing"]

            placed_components = placer_grid_cell_refs(
                components, cols, rows, dx, dy, x0=x0, y0=y0
            )

        # Place components within a cell having the DOE name
        placed_doe = gf.Component()
        placed_doe.add(placed_components)
        placed_doe.name = doe_name
        placed_does[doe_name] = placed_doe

        # # Write the json and md metadata / report
        # write_doe_metadata(
        # doe_name=doe_name,
        # cells = placed_components,
        # list_settings=list_settings,
        # flag_write_component=False,
        # description=description,
        # test=test,
        # analysis=analysis
        # )

        component_grid.add_ref(placed_doe)

    return component_grid


def build_components(
    component_type: str,
    list_settings: List[Dict[str, Union[float, int]]],
    component_factory: Dict[str, Callable] = factory,
) -> List[Component]:
    components = []

    # If no settings passed, generate a single component with defaults
    if not list_settings:
        component_function = component_factory[component_type]
        component = component_function()
        components += [component]
        return components

    for settings in list_settings:
        component_function = component_factory[component_type]
        component = component_function(**settings)
        components += [component]

    return components


if __name__ == "__main__":
    pass



import toolz

import gdsfactory as gf


def test_metadata_export_partial():
    straight_wide = gf.partial(gf.c.straight, width=2)
    c = gf.c.mzi(straight=straight_wide)
    d = c.to_dict_config()
    assert d.info.full.straight.width == 2
    assert d.info.full.straight.function == "straight"


def test_metadata_export_function():
    gf.clear_cache()
    c = gf.c.mzi()
    d = c.to_dict_config()
    assert d.info.full.straight.function == "straight"


def test_metadata_export_compose():
    straight_wide = toolz.compose(gf.c.extend_ports, gf.c.straight)
    c = gf.c.mzi(straight=straight_wide)
    d = c.to_dict_config()
    assert d.info.full.straight[0]["function"] == "straight"
    assert d.info.full.straight[1]["function"] == "extend_ports"


if __name__ == "__main__":
    test_metadata_export_compose()

    # test_metadata_export_function()
    # c = gf.c.mzi()
    # d = c.to_dict_config()
    # print(d.info.full.straight.function)

    # test_metadata_export_partial()
    # from gdsfactory.cell import clean_name
    # from gdsfactory.component import clean_dict
    # import inspect

    # straight_wide = gf.partial(gf.c.straight, width=2)
    # func = gf.c.mzi
    # func = gf.partial(func, straight=straight_wide)
    # c = gf.c.mzi(straight=straight_wide)
    # sig = inspect.signature(func)
    # default = {p.name: p.default for p in sig.parameters.values()}
    # full = default.copy()
    # clean_dict(full)

    # straight_wide = gf.partial(gf.c.straight, width=2)

    straight_wide = toolz.compose(gf.c.extend_ports, gf.c.straight)
    c = gf.c.mzi(straight=straight_wide)
    d = c.to_dict_config()
    print(d.info.full.straight)

    # df = d.info.full
    # sf = df.straight
    # print(sf)



import dataclasses
from typing import Callable, Dict, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from meep import mpb
from pydantic import BaseModel


@dataclasses.dataclass
class Mode:
    mode_number: int
    wavelength: float
    neff: float
    ng: Optional[float] = None
    fraction_te: Optional[float] = None
    fraction_tm: Optional[float] = None
    E: Optional[np.ndarray] = None
    H: Optional[np.ndarray] = None
    eps: Optional[np.ndarray] = None

    def __repr__(self):
        return f"Mode{self.mode_number}"

    def plot_eps(
        self,
        cmap: str = "viridis",
        origin="lower",
        logscale: bool = False,
        show: bool = True,
    ):
        """plot index profle"""
        plt.imshow(
            self.eps ** 0.5,
            cmap=cmap,
            origin=origin,
            aspect="auto",
        )
        plt.title("index profile")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_e(
        self,
        cmap: str = "viridis",
        origin="lower",
        logscale: bool = False,
        show: bool = True,
    ):
        E = self.E
        Eabs = np.sqrt(
            np.multiply(E[:, :, 0, 2], E[:, :, 0, 2])
            + np.multiply(E[:, :, 0, 1], E[:, :, 0, 1])
            + np.multiply(E[:, :, 0, 0], E[:, :, 0, 0])
        )
        ep = abs(Eabs)
        ep = 10 * np.log10(ep) if logscale else ep
        plt.imshow(
            ep.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
        )
        plt.title("Waveguide mode $|E|$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_ex(
        self,
        cmap: str = "viridis",
        origin="lower",
        logscale: bool = False,
        show: bool = True,
    ):
        ex = self.E[:, :, 0, 2]
        ex = 10 * np.log10(np.abs(ex)) if logscale else np.real(ex)
        plt.imshow(
            ex.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
        )
        plt.title("Waveguide mode $|E_x|$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_ey(
        self,
        cmap: str = "viridis",
        origin="lower",
        logscale: bool = False,
        show: bool = True,
    ):
        ey = self.E[:, :, 0, 1]
        ey = 10 * np.log10(np.abs(ey)) if logscale else np.real(ey)
        plt.imshow(
            ey.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
        )
        plt.title("Waveguide mode $|E_y|$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_ez(
        self,
        cmap: str = "viridis",
        origin="lower",
        logscale: bool = False,
        show: bool = True,
    ):
        E = self.E
        ez = abs(E[:, :, 0, 0])
        ez = 10 * np.log10(ez) if logscale else ez
        plt.imshow(
            ez.T,
            cmap=cmap,
            origin=origin,
            aspect="auto",
        )
        plt.title("Waveguide mode $|E_z|$")
        plt.ylabel("z-axis")
        plt.xlabel("y-axis")
        plt.colorbar()
        if show:
            plt.show()

    def plot_e_all(
        self,
        cmap: str = "viridis",
        origin="lower",
        logscale: bool = False,
        show: bool = True,
    ):
        plt.subplot(2, 3, 1)
        self.plot_e(show=False)

        plt.subplot(2, 3, 2)
        self.plot_ey(show=False)

        plt.subplot(2, 3, 3)
        self.plot_ex(show=False)

        plt.subplot(2, 3, 4)
        self.plot_ez(show=False)

        plt.subplot(2, 3, 5)
        self.plot_eps(show=False)

        plt.show()


@dataclasses.dataclass
class Waveguide:
    """
    Args:
        wg_width: float
        wg_thickness: float
        slab_thickness: float
        ncore: float = 3.47
        nclad: float = 1.44
        sy: float
        sz: float
        res: int
        nmodes: int
        modes: Dict[Mode]

    """

    wg_width: float = 0.45
    wg_thickness: float = 0.22
    slab_thickness: float = 0
    ncore: float = 3.47
    nclad: float = 1.44
    sy: float = 2.0
    sz: float = 2.0
    res: int = 32
    nmodes: int = 4
    modes: Optional[Dict[int, Mode]] = None


class WavelengthSweep(BaseModel):
    wavelength: List[float]
    neff: Dict[int, List[float]]
    ng: Dict[int, List[float]]


ModeSolverFactory = Callable[..., mpb.ModeSolver]
ModeSolverOrFactory = Union[mpb.ModeSolver, ModeSolverFactory]


if __name__ == "__main__":
    # import gdsfactory.simulation.modes as gm
    # m = gm.find_modes()
    # m[1].plot_e_all()
    w = Waveguide()



import phidl.geometry as pg

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.geometry.boolean import boolean
from gdsfactory.types import Int2, Layer, Union


@gf.cell
def invert(
    elements,
    border: float = 10.0,
    precision: float = 1e-4,
    num_divisions: Union[int, Int2] = (1, 1),
    max_points: int = 4000,
    layer: Layer = (1, 0),
):
    """Creates an inverted version of the input shapes with an additional
    border around the edges. adapted from phidl.geometry.invert

    Args:
        elements : Component(/Reference), list of Component(/Reference), or Polygon
            A Component containing the polygons to invert.
        border : int or float
            Size of the border around the inverted shape (border value is the
            distance from the edges of the boundary box defining the inverted
            shape to the border, and is applied to all 4 sides of the shape).
        precision : float
            Desired precision for rounding vertex coordinates.
        num_divisions : array-like[2] of int
            The number of divisions with which the geometry is divided into
            multiple rectangular regions. This allows for each region to be
            processed sequentially, which is more computationally efficient.
        max_points : int
            The maximum number of vertices within the resulting polygon.
        layer : int, array-like[2], or set
            Specific layer(s) to put polygon geometry on.

    Returns
        D: A Component containing the inverted version of the input shape(s) and the
        corresponding border(s).
    """
    Temp = Component()
    if type(elements) is not list:
        elements = [elements]
    for e in elements:
        if isinstance(e, Component):
            Temp.add_ref(e)
        else:
            Temp.add(e)
    gds_layer, gds_datatype = pg._parse_layer(layer)

    # Build the rectangle around the Component D
    R = gf.c.rectangle(
        size=(Temp.xsize + 2 * border, Temp.ysize + 2 * border), centered=True
    )
    R.center = Temp.center
    D = boolean(
        A=R,
        B=Temp,
        operation="A-B",
        precision=precision,
        num_divisions=num_divisions,
        max_points=max_points,
        layer=layer,
    )
    return D


def test_invert():
    e1 = gf.components.ellipse(radii=(6, 6)).move((10, 10))
    c = invert(e1)
    assert int(c.area()) == 910


if __name__ == "__main__":
    e1 = gf.components.ellipse(radii=(6, 6)).move((10, 10))
    c = invert(e1)
    c.show()



import inspect
from typing import Tuple

import gdsfactory as gf
from gdsfactory.add_labels import get_input_label
from gdsfactory.cell import cell
from gdsfactory.component import Component, ComponentReference
from gdsfactory.components.grating_coupler_elliptical_trenches import grating_coupler_te
from gdsfactory.cross_section import strip
from gdsfactory.port import Port
from gdsfactory.routing.get_route import get_route
from gdsfactory.routing.manhattan import round_corners
from gdsfactory.types import ComponentFactory, CrossSectionFactory


def connect_loopback(
    port0: Port,
    port1: Port,
    a: float,
    b: float,
    R: float,
    y_bot_align_route: float,
    cross_section: CrossSectionFactory = strip,
    **kwargs
) -> ComponentReference:
    p0 = port0.position
    p1 = port1.position
    points = [
        p0,
        p0 + (0, a),
        p0 + (b, a),
        p0 + (b, y_bot_align_route),
        p1 + (-b, y_bot_align_route),
        p1 + (-b, a),
        p1 + (0, a),
        p1,
    ]

    bend90 = gf.components.bend_euler(radius=R, cross_section=cross_section, **kwargs)
    return round_corners(
        points=points, bend=bend90, straight=gf.components.straight, **kwargs
    ).references


@cell
def loss_deembedding_ch13_24(
    pitch: float = 127.0,
    R: float = 10.0,
    grating_coupler_factory: ComponentFactory = grating_coupler_te,
    input_port_indexes: Tuple[int, ...] = (0, 1),
    cross_section: CrossSectionFactory = strip,
    **kwargs
) -> Component:

    gc = grating_coupler_factory()
    c = gf.Component()
    dx = pitch
    gcs = [gc.ref(position=(i * dx, 0), port_id="o1", rotation=-90) for i in range(4)]

    gc_ports = [g.ports["o1"] for g in gcs]
    c.add(gcs)

    c.add(
        get_route(
            gc_ports[0],
            gc_ports[2],
            start_straight_length=40.0,
            taper=None,
            cross_section=cross_section,
            **kwargs
        ).references
    )

    gsi = gc.size_info
    p1 = gc_ports[1]
    p3 = gc_ports[3]
    a = R + 5.0  # 0.5
    b = max(2 * a, pitch / 2)
    y_bot_align_route = -gsi.width - 5.0

    c.add(
        connect_loopback(
            p1, p3, a, b, R, y_bot_align_route, cross_section=cross_section, **kwargs
        )
    )
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.position = gc_ports[index].position
        c.add(label)

    return c


@cell
def loss_deembedding_ch12_34(
    pitch: float = 127.0,
    R: float = 10.0,
    grating_coupler_factory: ComponentFactory = grating_coupler_te,
    input_port_indexes: Tuple[int, ...] = (0, 2),
    **kwargs
) -> Component:
    gc = grating_coupler_factory()

    c = gf.Component()
    dx = pitch
    gcs = [gc.ref(position=(i * dx, 0), port_id="o1", rotation=-90) for i in range(4)]

    gc_ports = [g.ports["o1"] for g in gcs]
    c.add(gcs)

    c.add(
        get_route(
            gc_ports[0], gc_ports[1], start_straight_length=40.0, taper=None, **kwargs
        ).references
    )
    c.add(
        get_route(
            gc_ports[2], gc_ports[3], start_straight_length=40.0, taper=None, **kwargs
        ).references
    )
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.position = gc_ports[index].position
        c.add(label)
    return c


@cell
def loss_deembedding_ch14_23(
    pitch: float = 127.0,
    R: float = 10.0,
    grating_coupler_factory: ComponentFactory = grating_coupler_te,
    input_port_indexes: Tuple[int, ...] = (0, 1),
    **kwargs
) -> Component:
    gc = grating_coupler_factory()

    c = gf.Component()
    dx = pitch
    gcs = [gc.ref(position=(i * dx, 0), port_id="o1", rotation=-90) for i in range(4)]

    gc_ports = [g.ports["o1"] for g in gcs]
    c.add(gcs)

    c.add(
        get_route(
            gc_ports[0], gc_ports[3], start_straight_length=40.0, taper=None, **kwargs
        ).references
    )
    c.add(
        get_route(
            gc_ports[1], gc_ports[2], start_straight_length=30.0, taper=None, **kwargs
        ).references
    )
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.position = gc_ports[index].position
        c.add(label)
    return c


@cell
def grating_coupler_loss(
    pitch: float = 127.0,
    grating_coupler_factory: ComponentFactory = grating_coupler_te,
    **kwargs
) -> Component:
    """

    Args:
        pitch: grating_coupler_pitch
        grating_coupler_factory: function
        **kwargs: cross_section settings

    """
    c = gf.Component()
    _c1 = loss_deembedding_ch13_24(
        grating_coupler_factory=grating_coupler_factory, **kwargs
    )
    _c2 = loss_deembedding_ch14_23(
        grating_coupler_factory=grating_coupler_factory, **kwargs
    )
    _c3 = loss_deembedding_ch12_34(
        grating_coupler_factory=grating_coupler_factory, **kwargs
    )
    c.add_ref(_c1)
    c2 = c.add_ref(_c2)
    c3 = c.add_ref(_c3)
    c2.movex(pitch * 4)
    c3.movex(pitch * 8)

    return c


if __name__ == "__main__":
    # c = loss_deembedding_ch14_23()
    # c = loss_deembedding_ch12_34()
    # c = loss_deembedding_ch13_24()
    c = grating_coupler_loss(layer=(2, 0))
    c.show()



"""Define names, clean values for names.
"""
import functools
import hashlib
import inspect
from typing import Any, Iterable

import numpy as np
import pydantic
import toolz
from phidl import Device, Port
from phidl.device_layout import Path as PathPhidl

from gdsfactory.hash_points import hash_points
from gdsfactory.snap import snap_to_grid

MAX_NAME_LENGTH = 32


@pydantic.validate_arguments
def get_name_short(name: str, max_name_length=MAX_NAME_LENGTH) -> str:
    """Returns a short name."""
    if len(name) > max_name_length:
        name_hash = hashlib.md5(name.encode()).hexdigest()[:8]
        name = f"{name[:(max_name_length - 9)]}_{name_hash}"
    return name


def join_first_letters(name: str) -> str:
    """Join the first letter of a name separated with underscores.

    taper_length -> TL
    """
    return "".join([x[0] for x in name.split("_") if x])


# replace function_name prefix for some components
component_type_to_name = dict(phidl="phidl")


def get_component_name(component_type: str, *args, **kwargs) -> str:
    """Returns concatenated kwargs Key_Value."""
    name = component_type
    name += "_".join([clean_value(a) for a in args])
    for k, v in component_type_to_name.items():
        name = name.replace(k, v)
    if kwargs:
        name += "_" + dict2name(**kwargs)
    return name


def dict2hash(**kwargs) -> str:
    ignore_from_name = kwargs.pop("ignore_from_name", [])
    h = hashlib.sha256()
    for key in sorted(kwargs):
        if key not in ignore_from_name:
            value = kwargs[key]
            value = clean_value(value)
            h.update(f"{key}{value}".encode())
    return h.hexdigest()


def dict2name(prefix: str = "", **kwargs) -> str:
    """Returns name from a dict."""
    ignore_from_name = kwargs.pop("ignore_from_name", [])
    kv = []
    kwargs = kwargs.copy()
    kwargs.pop("layer_to_inclusion", "")

    for key in sorted(kwargs):
        if key not in ignore_from_name and isinstance(key, str):
            value = kwargs[key]
            # key = join_first_letters(key).upper()
            if value is not None:
                kv += [f"{key}{clean_value(value)}"]
    label = prefix + "_".join(kv)
    return clean_name(label)


def assert_first_letters_are_different(**kwargs):
    """Assert that the first letters for each key are different.

    Avoids name collisions of different args that start with the same first letter.
    """
    first_letters = [join_first_letters(k) for k in kwargs.keys()]
    if not len(set(first_letters)) == len(first_letters):
        raise ValueError(
            f"Possible name collision! {kwargs.keys()} repeats first letters {first_letters}",
            "you can separate your arguments with underscores",
            " (delta_length -> DL, delta_width -> DW",
        )


def print_first_letters_warning(**kwargs):
    """Prints kwargs that have same cell."""
    first_letters = [join_first_letters(k) for k in kwargs.keys()]
    if not len(set(first_letters)) == len(first_letters):
        print(
            f"Possible name collision! {kwargs.keys()} "
            f"repeats first letters {first_letters}"
            "you can separate your arguments with underscores"
            " (delta_length -> DL, delta_width -> DW"
        )


def clean_name(name: str) -> str:
    """Return a string with correct characters for a cell name.

    [a-zA-Z0-9]

    FIXME: only a few characters are currently replaced.
        This function has been updated only on case-by-case basis
    """
    replace_map = {
        " ": "_",
        "!": "",
        "?": "",
        "#": "_",
        "%": "_",
        "(": "",
        ")": "",
        "*": "_",
        ",": "_",
        "-": "m",
        ".": "p",
        "/": "_",
        ":": "_",
        "=": "",
        "@": "_",
        "[": "",
        "]": "",
    }
    for k, v in list(replace_map.items()):
        name = name.replace(k, v)
    return name


def clean_value(value: Any) -> str:
    """returns more readable value (integer)
    if number is < 1:
        returns number units in nm (integer)

    units are in um by default. Therefore when we multiply by 1e3 we get nm.
    """

    if isinstance(value, int):
        value = str(value)
    elif isinstance(value, (float, np.float64)):
        if 1 > value > 1e-3:
            value = f"{int(value*1e3)}n"
        elif int(value) == value:
            value = str(int(value))
        elif 1e-6 < value < 1e-3:
            value = f"{snap_to_grid(value*1e6)}u"
        elif 1e-9 < value < 1e-6:
            value = f"{snap_to_grid(value*1e9)}n"
        elif 1e-12 < value < 1e-9:
            value = f"{snap_to_grid(value*1e12)}p"
        else:  # Any unit < 1pm will disappear
            value = str(snap_to_grid(value)).replace(".", "p")
    elif isinstance(value, Device):
        value = clean_name(value.name)
    elif isinstance(value, str):
        value = value.strip()
    elif isinstance(value, dict):
        value = dict2name(**value)
        # value = [f"{k}={v!r}" for k, v in value.items()]
    elif isinstance(value, Port):
        value = f"{value.name}_{value.width}_{value.x}_{value.y}"
    elif isinstance(value, PathPhidl):
        value = f"path_{hash_points(value.points)}"
    elif (
        isinstance(value, object)
        and hasattr(value, "name")
        and isinstance(value.name, str)
    ):
        value = clean_name(value.name)
    elif callable(value) and isinstance(value, functools.partial):
        sig = inspect.signature(value.func)
        args_as_kwargs = dict(zip(sig.parameters.keys(), value.args))
        args_as_kwargs.update(**value.keywords)
        value = value.func.__name__ + dict2name(**args_as_kwargs)
    elif callable(value) and isinstance(value, toolz.functoolz.Compose):
        value = "_".join(
            [clean_value(v) for v in value.funcs] + [clean_value(value.first)]
        )
    elif callable(value) and hasattr(value, "__name__"):
        value = value.__name__
    elif hasattr(value, "get_name"):
        value = value.get_name()
    elif isinstance(value, Iterable):
        value = "_".join(clean_value(v) for v in value)

    return str(value)


def test_clean_value() -> None:
    assert clean_value(0.5) == "500n"
    assert clean_value(5) == "5"
    assert clean_value(5.0) == "5"
    assert clean_value(11.001) == "11p001"


def test_clean_name() -> None:
    assert clean_name("wg(:_=_2852") == "wg___2852"


if __name__ == "__main__":
    # test_cell()
    test_clean_value()
    import gdsfactory as gf

    # print(clean_value(gf.components.straight))
    # c = gf.components.straight(polarization="TMeraer")
    # print(c.settings["polarization"])
    # print(clean_value(11.001))
    # layers_cladding = (gf.LAYER.WGCLAD, gf.LAYER.NO_TILE_SI)
    # layers_cladding = (gf.LAYER.WGCLAD,)
    c = gf.components.straight(length=10)
    c = gf.components.straight(length=10)

    # print(c.name)
    # print(c)
    # c.show()

    # print(clean_name("Waveguidenol1_(:_=_2852"))
    # print(clean_value(1.2))
    # print(clean_value(0.2))
    # print(clean_value([1, [2.4324324, 3]]))
    # print(clean_value([1, 2.4324324, 3]))
    # print(clean_value((0.001, 24)))
    # print(clean_value({"a": 1, "b": 2}))



import gdsfactory as gf


def test_rotate():
    c1 = gf.c.straight()
    c1r = c1.rotate()

    c2 = gf.c.straight()
    c2r = c2.rotate()

    assert c1.uid == c2.uid
    assert c1r.uid == c2r.uid


if __name__ == "__main__":
    c1 = gf.c.straight()
    c1r = c1.rotate()

    c2 = gf.c.straight()
    c2r = c2.rotate()

    assert c1.uid == c2.uid
    assert c1r.uid == c2r.uid
    c2r.show()



from typing import Tuple, Union

import numpy as np


def is_on_grid(x: float, nm: int = 1) -> bool:
    return np.isclose(snap_to_grid(x, nm=nm), x)


def assert_on_1nm_grid(x: float) -> None:
    x_grid = snap_to_grid(x)
    assert np.isclose(x_grid, x), f"{x} needs to be on 1nm grid, try {x_grid}"


def assert_on_2nm_grid(x: float) -> None:
    x_grid = snap_to_2nm_grid(x)
    assert np.isclose(x_grid, x), f"{x} needs to be on 2nm grid, try {x_grid}"


def snap_to_grid(
    x: Union[float, Tuple[float, ...], np.array], nm: int = 1
) -> Union[float, np.array, Tuple[float, ...]]:
    y = nm * np.round(np.array(x, dtype=float) * 1e3 / nm) / 1e3
    if isinstance(x, tuple):
        return tuple(y)
    elif type(x) in [int, float, str, np.float64]:
        return float(y)
    return y


def snap_to_2nm_grid(x: float) -> float:
    return snap_to_grid(x, nm=2)


def snap_to_5nm_grid(x: float) -> float:
    return snap_to_grid(x, nm=5)


if __name__ == "__main__":
    print(snap_to_grid(1.1e-3))
    # print(snap_to_2nm_grid(1.1e-3))
    # print(snap_to_2nm_grid(3.1e-3))

    # print(on_1nm_grid(1.1e-3))
    # print(on_1nm_grid(1e-3))

    # print(on_2nm_grid(1.1e-3))
    # print(on_2nm_grid(1e-3))
    # print(on_2nm_grid(2e-3))



import gdsfactory as gf
from gdsfactory.components.contact import contact
from gdsfactory.components.rectangle import rectangle
from gdsfactory.snap import snap_to_grid

big_square = gf.partial(rectangle, size=(1300, 2600))


@gf.cell
def seal_ring(
    component: gf.types.ComponentOrFactory = big_square,
    seal: gf.types.ComponentFactory = contact,
    width: float = 10,
    padding: float = 10.0,
    with_north: bool = True,
    with_south: bool = True,
    with_east: bool = True,
    with_west: bool = True,
) -> gf.Component:
    """Returns a continuous seal ring boundary at the chip/die
    seal rings are useful to prevents cracks from spreading
    you can connect it to ground

    Args:
        component: to add seal ring around
        seal: function for the seal
        width: of the seal
        padding: from component to seal
        with_north: includes seal
        with_south: includes seal
        with_east: includes seal
        with_west: includes seal

    """
    c = gf.Component()
    component = component() if callable(component) else component
    size = component.size
    sx, sy = size
    sx = gf.snap.snap_to_grid(sx, nm=2)
    sy = gf.snap.snap_to_grid(sy, nm=2)

    # north south
    size_north_south = (sx + 2 * padding + 2 * width, width)
    size_east_west = (width, sy + 2 * padding)

    if with_north:
        north = c << seal(size=size_north_south)
        north.ymin = snap_to_grid(component.ymax + padding)
        north.x = snap_to_grid(component.x)

    if with_south:
        south = c << seal(size=size_north_south)
        south.ymax = snap_to_grid(component.ymin - padding)
        south.x = snap_to_grid(component.x)

    if with_east:
        east = c << seal(size=size_east_west)
        east.xmin = snap_to_grid(component.xmax + padding)
        east.y = snap_to_grid(component.y)

    if with_west:
        west = c << seal(size=size_east_west)
        west.xmax = snap_to_grid(component.xmin - padding)
        west.y = snap_to_grid(component.y)

    return c


if __name__ == "__main__":
    c = big_square()
    c << seal_ring(with_south=False)
    c.show()



import pathlib
from typing import Tuple

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.import_gds import import_gds
from gdsfactory.types import ComponentOrPath, PathType


@cell
def from_gdspaths(cells: Tuple[ComponentOrPath, ...]) -> Component:
    """Combine all GDS files or gf.components into a gf.component.

    Args:
        cells: List of gdspaths or Components
    """
    component = Component()

    for c in cells:
        if not isinstance(c, Component):
            c = import_gds(c)
        component << c

    return component


def gdsdir(dirpath: PathType) -> Component:
    """Merges GDS cells from a directory into a single Component"""
    dirpath = pathlib.Path(dirpath)
    return from_gdspaths(dirpath.glob("*.gds"))


if __name__ == "__main__":
    from gdsfactory.config import diff_path

    # c = gdspaths([gf.components.straight(), gf.components.bend_circular()])
    # leave these two lines to end up tests showing the diff
    c = from_gdspaths(diff_path.glob("*.gds"))
    c.show()



"""All functions return a component so they are easy to pipe and compose."""
from functools import lru_cache

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.port import auto_rename_ports
from gdsfactory.types import ComponentOrFactory, Float2, Optional

cache = lru_cache(maxsize=None)


def add_port(component: Component, **kwargs) -> Component:
    """Returns Component with a new port."""
    component.add_port(**kwargs)
    return component


@cell
def rotate(
    component: ComponentOrFactory,
    angle: int = 90,
) -> Component:
    """Returns rotated component inside a new component.

    Most times you just need to place a reference and rotate it.
    This rotate function just encapsulates the rotated reference into a new component.

    Args:
        component:
        angle: in degrees
    """
    component = component() if callable(component) else component
    component_new = Component()
    component_new.component = component
    ref = component_new.add_ref(component)
    ref.rotate(angle)
    component_new.add_ports(ref.ports)
    component_new.copy_child_info(component)
    return component_new


@cell
def mirror(component: Component, p1: Float2 = (0, 1), p2: Float2 = (0, 0)) -> Component:
    """Returns mirrored component inside a new component.

    Args:
        p1: first point to define mirror axis
        p2: second point to define mirror axis
    """
    component_new = Component()
    component_new.component = component
    ref = component_new.add_ref(component)
    ref.mirror(p1=p1, p2=p2)
    component_new.add_ports(ref.ports)
    component_new.copy_child_info(component)
    return component_new


@cell
def move(
    component: Component,
    origin=(0, 0),
    destination=None,
    axis: Optional[str] = None,
) -> Component:
    """Returns a container that contains a reference to the original component."""
    component_new = Component()
    component_new.component = component
    ref = component_new.add_ref(component)
    ref.move(origin=origin, destination=destination, axis=axis)
    component_new.add_ports(ref.ports)
    component_new.copy_child_info(component)
    return component_new


def move_port_to_zero(component: Component, port_name: str = "o1"):
    """Returns a container that contains a reference to the original component.
    where the new component has port_name in (0, 0)
    """
    if port_name not in component.ports:
        raise ValueError(
            f"port_name = {port_name} not in {list(component.ports.keys())}"
        )
    return move(component, -component.ports[port_name].midpoint)


def update_info(component: Component, **kwargs) -> Component:
    """Returns Component with updated info."""
    component.info.update(**kwargs)
    return component


__all__ = (
    "cache",
    "add_port",
    "rotate",
    "auto_rename_ports",
    "move",
    "move_port_to_zero",
    "update_info",
)


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.mmi1x2()
    cr = c.rotate()
    cr.show()

    cm = move(c, destination=(20, 20))
    cm.show()

    cm = mirror(c)
    cm.show()

    cm = c.mirror()
    cm.show()

    cm2 = move_port_to_zero(cm)
    cm2.show()

    # cr = rotate(component=c)
    # cr.show()
    # print(component_rotated)

    # component_rotated.pprint
    # component_netlist = component.get_netlist()
    # component.pprint_netlist()



""" adiabatic tapers from CSV files
"""
import pathlib
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd

import gdsfactory as gf
from gdsfactory.component import Component

data_path = pathlib.Path(__file__).parent / "csv_data"


@gf.cell
def taper_from_csv(
    filepath: Path = data_path / "taper_strip_0p5_3_36.csv",
    layer: Tuple[int, int] = (1, 0),
    layer_cladding: Tuple[int, int] = gf.LAYER.WGCLAD,
    cladding_offset: float = 3.0,
    **kwargs
) -> Component:
    """

    Args:
        filepath: for CSV file
        layer
        layer_cladding:
        cladding_offset
        kwargs: kwargs will be ignored

    """
    taper_data = pd.read_csv(filepath)
    xs = taper_data["x"].values * 1e6
    ys = np.round(taper_data["width"].values * 1e6 / 2.0, 3)
    ys_trench = ys + cladding_offset

    c = gf.Component()
    c.add_polygon(list(zip(xs, ys)) + list(zip(xs, -ys))[::-1], layer=layer)
    c.add_polygon(
        list(zip(xs, ys_trench)) + list(zip(xs, -ys_trench))[::-1], layer=layer_cladding
    )

    c.add_port(
        name="o1",
        midpoint=(xs[0], 0),
        width=2 * ys[0],
        orientation=180,
    )
    c.add_port(
        name="o2",
        midpoint=(xs[-1], 0),
        width=2 * ys[-1],
        orientation=0,
    )
    return c


@gf.cell
def taper_0p5_to_3_l36(**kwargs) -> Component:
    filepath = data_path / "taper_strip_0p5_3_36.csv"
    return taper_from_csv(filepath=filepath, **kwargs)


@gf.cell
def taper_w10_l100(**kwargs):
    filepath = data_path / "taper_strip_0p5_10_100.csv"
    return taper_from_csv(filepath=filepath, **kwargs)


@gf.cell
def taper_w10_l150(**kwargs):
    filepath = data_path / "taper_strip_0p5_10_150.csv"
    return taper_from_csv(filepath=filepath, **kwargs)


@gf.cell
def taper_w10_l200(**kwargs):
    filepath = data_path / "taper_strip_0p5_10_200.csv"
    return taper_from_csv(filepath=filepath, **kwargs)


@gf.cell
def taper_w11_l200(**kwargs):
    filepath = data_path / "taper_strip_0p5_11_200.csv"
    return taper_from_csv(filepath=filepath, **kwargs)


@gf.cell
def taper_w12_l200(**kwargs):
    filepath = data_path / "taper_strip_0p5_12_200.csv"
    return taper_from_csv(filepath=filepath, **kwargs)


if __name__ == "__main__":
    c = taper_0p5_to_3_l36()
    # c = taper_w10_l100()
    # c = taper_w11_l200()
    c.show()



"""GDS regression test. Adapted from lytest.

TODO: adapt it into pytest_regressions

from pytest_regressions.file_regression import FileRegressionFixture

class GdsRegressionFixture(FileRegressionFixture):
    def check(self,
        contents,
        extension=".gds",
        basename=None,
        fullpath=None,
        binary=False,
        obtained_filename=None,
        check_fn=None,
            ):
        try:
            difftest(c)

"""
import filecmp
import pathlib
from typing import Optional

from lytest.kdb_xor import GeometryDifference, run_xor

from gdsfactory.component import Component
from gdsfactory.gdsdiff.gdsdiff import gdsdiff

cwd = pathlib.Path.cwd()


def difftest(
    component: Component,
    test_name: Optional[str] = None,
    xor: bool = False,
    dirpath: pathlib.Path = cwd,
) -> None:
    """Avoids GDS regressions tests on the GeometryDifference.
    Runs an XOR over a component and makes boolean comparison with a GDS reference.
    If it runs for the fist time it just stores the GDS reference.
    raises GeometryDifference if there are differences and show differences in klayout.


    Args:
        component:
        test_name: used to store the GDS file
        xor: runs xor if there is difference
        dirpath: defaults to cwd refers to where the test is being invoked
    """

    # containers function_name is different from component.name
    # we store the container with a different name from original component
    test_name = test_name or (
        f"{component.function_name}_{component.name}"
        if hasattr(component, "function_name")
        and component.name != component.function_name
        else f"{component.name}"
    )
    filename = f"{test_name}.gds"
    ref_file = dirpath / "gds_ref" / filename
    run_file = dirpath / "gds_run" / filename
    diff_file = dirpath / "gds_diff" / filename

    component.write_gds(gdspath=run_file)

    if not ref_file.exists():
        print(f"Creating GDS reference for {component.name} in {ref_file}")
        component.write_gds(gdspath=ref_file)
        return

    if filecmp.cmp(ref_file, run_file, shallow=False):
        return

    # component_reference = gf.import_gds(ref_file)
    # if component.hash_geometry() == component_reference.hash_geometry():
    #     print("same hash")
    #     return

    try:
        run_xor(str(ref_file), str(run_file), tolerance=1, verbose=False)
    except GeometryDifference as error:
        print()
        print(error)
        diff = gdsdiff(ref_file, run_file, name=test_name, xor=xor)
        diff.write_gds(diff_file)
        diff.show(show_ports=False)
        print(
            "\n"
            + f"`{filename}` changed from reference {ref_file}\n"
            + "You can check the differences in Klayout GUI\n"
            + "For a detailed XOR you can run\n"
            + f"gf gds diff --xor {ref_file} {run_file}\n"
        )

        try:
            val = input(
                "Would you like to save current GDS as the new reference? [y/n] "
            )
            if val.upper().startswith("Y"):
                print(f"rm {ref_file}")
                ref_file.unlink()
            raise
        except OSError as exc:
            raise GeometryDifference(
                "\n"
                + f"`{filename}` changed from reference {ref_file}\n"
                + "To step over each error you can run `pytest -s`\n"
                + "So you can check the differences in Klayout GUI\n"
            ) from exc


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.straight()
    difftest(c)
    # test_component(c, None, None)



"""Tidy3D """

import concurrent.futures
import hashlib
import json
import pathlib
from typing import Awaitable, Optional

import tidy3d as td
from tidy3d import web

from gdsfactory.config import PATH, logger

_executor = concurrent.futures.ThreadPoolExecutor()


def _export_simulation(
    sim: td.Simulation,
    task_name: Optional[str] = None,
    folder_name: str = "default",
    draft: bool = False,
) -> int:
    """Exports simulation to web and returns task_id.

    Args:
        sim: simulation object
        task_name:
        folder_name: Server folder to hold the task.
        draft: If ``True``, the project will be submitted but not run.
            It can then be visualized in the web UI and run from there when needed.

    """
    project = web.new_project(
        sim.export(), task_name=task_name, folder_name=folder_name, draft=draft
    )
    task_id = project["taskId"]
    logger.info(f"submitting {task_id}")
    return task_id


def get_sim_hash(sim: td.Simulation) -> str:
    """Returns simulation hash as the unique ID."""
    sim_str = json.dumps(sim.export())
    sim_hash = hashlib.md5(sim_str.encode()).hexdigest()
    return sim_hash


def load_results(
    sim: td.Simulation, target: pathlib.Path, task_id: Optional[str] = None
) -> td.Simulation:
    """Load results from HDF5 file. Returns a simulation that includes results.

    Args:
        sim: Simulation
        target: path
        task_id: Optional task Id
    """

    if task_id:
        web.monitor_project(task_id)
        src = "monitor_data.hdf5"
        web.download_results_file(task_id=task_id, src=src, target=str(target))

    sim.load_results(target)
    return sim


def run_simulation(sim: td.Simulation) -> Awaitable[td.Simulation]:
    """Returns a simulation with simulation results

    Only submits simulation if results not found locally or remotely.

    First tries to load simulation results from disk.
    Then it tries to load them from the server storage.
    Finally, only submits simulation if not found


    .. code::
        import gdsfactory.simulation.tidy3d as gm

        component = gf.components.straight(length=3)
        sim = gm.get_simulation(component=component)
        sim = run_simulation(sim).result()

    """
    td.logging_level("error")
    sim_hash = get_sim_hash(sim)
    sim_path = PATH.results_tidy3d / f"{sim_hash}.hdf5"
    logger.info(f"running simulation {sim_hash}")

    hash_to_id = {d["task_name"][:32]: d["task_id"] for d in web.get_last_projects()}
    target = PATH.results_tidy3d / f"{sim_hash}.hdf5"

    # Try from local storage
    if sim_path.exists():
        logger.info(f"{sim_path} found in local storage")
        sim = _executor.submit(load_results, sim, target)

    # Try from server storage
    elif sim_hash in hash_to_id:
        task_id = hash_to_id[sim_hash]
        sim = _executor.submit(load_results, sim, target, task_id)

    # Only submit if simulation not found
    else:
        task_id = _export_simulation(sim=sim, task_name=sim_hash)
        sim = _executor.submit(load_results, sim, target, task_id)
    return sim


if __name__ == "__main__":
    import gdsfactory as gf
    import gdsfactory.simulation.tidy3d as gm

    # for length in range(12, 13):
    #     component = gf.components.straight(length=length)
    #     sim = gm.get_simulation(component=component)
    #     s = run_simulation(sim, wait_to_complete=False)

    component = gf.components.straight(length=3)
    s = gm.get_simulation(component=component)
    r = run_simulation(s).result()



"""Compute and write Sparameters using Meep."""

import pathlib
from pathlib import Path
from typing import Dict, Optional, Tuple

import matplotlib.pyplot as plt
import meep as mp
import numpy as np
import pandas as pd
import pydantic

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.config import CONFIG
from gdsfactory.simulation.get_sparameters_path import get_sparameters_path
from gdsfactory.simulation.gmeep.get_simulation import (
    LAYER_TO_MATERIAL,
    LAYER_TO_THICKNESS,
    get_simulation,
)


@pydantic.validate_arguments
def get_sparameters1x1(
    component: Component,
    dirpath: Path = CONFIG["sparameters"],
    layer_to_thickness: Dict[Tuple[int, int], float] = LAYER_TO_THICKNESS,
    layer_to_material: Dict[Tuple[int, int], str] = LAYER_TO_MATERIAL,
    filepath: Optional[Path] = None,
    overwrite: bool = False,
    **settings,
) -> pd.DataFrame:
    """Compute Sparameters and writes them in CSV filepath.

    Args:
        component: to simulate.
        dirpath: directory to store Sparameters
        layer_to_thickness: GDS layer (int, int) to thickness
        layer_to_material: GDS layer (int, int) to material string ('Si', 'SiO2', ...)
        filepath: to store pandas Dataframe with Sparameters in CSV format
        overwrite: overwrites
        **settings: sim settings

    Returns:
        sparameters in a pandas Dataframe

    """
    filepath = filepath or get_sparameters_path(
        component=component,
        dirpath=dirpath,
        layer_to_material=layer_to_material,
        layer_to_thickness=layer_to_thickness,
    )
    filepath = pathlib.Path(filepath)
    if filepath.exists() and not overwrite:
        print(f"Simulation loaded from file: {filepath}")
        return pd.read_csv(filepath)

    sim_dict = get_simulation(
        component=component,
        layer_to_thickness=layer_to_thickness,
        layer_to_material=layer_to_material,
        **settings,
    )

    sim = sim_dict["sim"]
    monitors = sim_dict["monitors"]
    freqs = sim_dict["freqs"]
    field_monitor_point = sim_dict["field_monitor_point"]
    wavelengths = 1 / freqs
    port1 = sim_dict["port_source_name"]
    port2 = set(monitors.keys()) - set([port1])
    port2 = list(port2)[0]
    monitor1 = monitors[port1]
    monitor2 = monitors[port2]

    sim.run(
        until_after_sources=mp.stop_when_fields_decayed(
            dt=50, c=mp.Ez, pt=field_monitor_point, decay_by=1e-9
        )
    )
    # call this function every 50 time spes
    # look at simulation and measure component that we want to measure (Ez component)
    # when field_monitor_point decays below a certain 1e-9 field threshold

    # Calculate mode overlaps
    m_results = np.abs(sim.get_eigenmode_coefficients(monitor1, [1]).alpha)
    a = m_results[:, :, 0]  # forward wave
    source_fields = np.squeeze(a)
    t = sim.get_eigenmode_coefficients(monitor2, [1]).alpha[0, :, 0] / source_fields
    r = sim.get_eigenmode_coefficients(monitor1, [1]).alpha[0, :, 1] / source_fields

    s11 = r
    s12 = t

    s11a = np.unwrap(np.angle(s11))
    s12a = np.unwrap(np.angle(s12))
    s11m = np.abs(s11)
    s12m = np.abs(s12)

    df = pd.DataFrame(
        dict(
            wavelengths=wavelengths,
            freqs=freqs,
            s11m=s11m,
            s11a=s11a,
            s12m=s12m,
            s12a=s12a,
            s22m=s11m,
            s22a=s11a,
            s21m=s12m,
            s21a=s12a,
        )
    )
    print(f"transmission: {t}")

    return df


def plot_sparameters(df: pd.DataFrame) -> None:
    """Plot Sparameters from a Pandas DataFrame."""
    wavelengths = df["wavelengths"]
    for key in df.keys():
        if key.endswith("m"):
            plt.plot(
                wavelengths,
                df[key],
                "-o",
                label=key,
            )
    plt.ylabel("Power (dB)")
    plt.xlabel(r"Wavelength ($\mu$m)")
    plt.legend()
    plt.grid(True)


if __name__ == "__main__":

    c = gf.components.bend_circular(radius=2)
    c = gf.add_padding(c, default=0, bottom=2, right=2, layers=[(100, 0)])

    c = gf.components.mmi1x2()
    c = gf.add_padding(c, default=0, bottom=2, top=2, layers=[(100, 0)])

    c = gf.components.straight(length=2)
    c = gf.add_padding(c, default=0, bottom=2, top=2, layers=[(100, 0)])

    sim_dict = get_simulation(c, is_3d=False)
    df = get_sparameters1x1(c, overwrite=True)
    plot_sparameters(df)
    plt.show()
    print(df)



import numpy as np
import tidy3d as td

from gdsfactory.simulation.tidy3d.get_sparameters import get_sparameters


def get_coupling(sim: td.Simulation) -> np.ndarray:
    """Adapted from tidy3d examples
    https://simulation.cloud/docs/html/examples/ParameterScan.html

    Computes the transmission  and efficiency of a 2x2 coupler
    """
    S = get_sparameters(sim)
    split_ratio_top = S[0, 1] / (S[0, 1] + S[1, 1])
    split_ratio_bot = S[1, 1] / (S[0, 1] + S[1, 1])
    print(f"split ratio of {(split_ratio_top * 100):.2f}% to top port")
    print(f"split ratio of {(split_ratio_bot * 100):.2f}% to bot port")

    # efficiency = S[0, 1] + S[1, 1]
    # print(f"efficiency of {(efficiency * 100):.2f}% to transmission port")
    return S


if __name__ == "__main__":
    import gdsfactory as gf
    import gdsfactory.simulation.tidy3d as gm

    c = gf.components.coupler(gap=0.1, length=6.0)
    sim = gm.get_simulation(c)

    s = get_coupling(sim)



from gdsfactory.read.from_gds import from_gds
from gdsfactory.read.from_gdspaths import from_gdspaths, gdsdir
from gdsfactory.read.from_np import from_np
from gdsfactory.read.from_phidl import from_phidl
from gdsfactory.read.from_picwriter import from_picwriter
from gdsfactory.read.from_yaml import from_yaml

__all__ = [
    "from_phidl",
    "from_picwriter",
    "from_gds",
    "from_gdspaths",
    "gdsdir",
    "from_np",
    "from_yaml",
]



from gdsfactory.components.align import add_frame, align_wafer
from gdsfactory.components.array import array
from gdsfactory.components.array_with_fanout import (
    array_with_fanout,
    array_with_fanout_2d,
)
from gdsfactory.components.array_with_via import array_with_via, array_with_via_2d
from gdsfactory.components.awg import awg
from gdsfactory.components.bbox import bbox
from gdsfactory.components.bend_circular import bend_circular, bend_circular180
from gdsfactory.components.bend_circular_heater import bend_circular_heater
from gdsfactory.components.bend_euler import (
    bend_euler,
    bend_euler180,
    bend_euler_s,
    bend_straight_bend,
)
from gdsfactory.components.bend_port import bend_port
from gdsfactory.components.bend_s import bend_s
from gdsfactory.components.C import C
from gdsfactory.components.cavity import cavity
from gdsfactory.components.cdc import cdc
from gdsfactory.components.circle import circle
from gdsfactory.components.compass import compass
from gdsfactory.components.component_lattice import component_lattice
from gdsfactory.components.component_sequence import component_sequence
from gdsfactory.components.contact import contact, contact_heater_m3, contact_slab_m3
from gdsfactory.components.contact_slot import contact_slot, contact_slot_m1_m2
from gdsfactory.components.contact_with_offset import contact_with_offset
from gdsfactory.components.copy_layers import copy_layers
from gdsfactory.components.coupler import coupler
from gdsfactory.components.coupler90 import coupler90, coupler90circular
from gdsfactory.components.coupler90bend import coupler90bend
from gdsfactory.components.coupler_adiabatic import coupler_adiabatic
from gdsfactory.components.coupler_asymmetric import coupler_asymmetric
from gdsfactory.components.coupler_full import coupler_full
from gdsfactory.components.coupler_ring import coupler_ring
from gdsfactory.components.coupler_straight import coupler_straight
from gdsfactory.components.coupler_symmetric import coupler_symmetric
from gdsfactory.components.cross import cross
from gdsfactory.components.crossing_waveguide import (
    compensation_path,
    crossing,
    crossing45,
    crossing_arm,
    crossing_etched,
    crossing_from_taper,
)
from gdsfactory.components.cutback_bend import (
    cutback_bend,
    cutback_bend90,
    cutback_bend90circular,
    cutback_bend180,
    cutback_bend180circular,
    staircase,
)
from gdsfactory.components.cutback_component import (
    cutback_component,
    cutback_component_mirror,
)
from gdsfactory.components.dbr import dbr
from gdsfactory.components.dbr_tapered import dbr_tapered
from gdsfactory.components.delay_snake import delay_snake
from gdsfactory.components.delay_snake2 import delay_snake2, test_delay_snake2_length
from gdsfactory.components.delay_snake3 import delay_snake3, test_delay_snake3_length
from gdsfactory.components.dicing_lane import dicing_lane
from gdsfactory.components.die import die
from gdsfactory.components.die_bbox import big_square, die_bbox
from gdsfactory.components.disk import disk
from gdsfactory.components.ellipse import ellipse
from gdsfactory.components.extend_ports_list import extend_ports_list
from gdsfactory.components.extension import extend_port, extend_ports
from gdsfactory.components.fiber import fiber
from gdsfactory.components.fiber_array import fiber_array
from gdsfactory.components.grating_coupler_array import grating_coupler_array
from gdsfactory.components.grating_coupler_circular import (
    grating_coupler_circular,
    grating_coupler_circular_arbitrary,
)
from gdsfactory.components.grating_coupler_elliptical import (
    ellipse_arc,
    grating_coupler_elliptical,
    grating_coupler_elliptical_te,
    grating_coupler_elliptical_tm,
    grating_taper_points,
    grating_tooth_points,
)
from gdsfactory.components.grating_coupler_elliptical_arbitrary import (
    grating_coupler_elliptical_arbitrary,
)
from gdsfactory.components.grating_coupler_elliptical_lumerical import (
    grating_coupler_elliptical_lumerical,
)
from gdsfactory.components.grating_coupler_elliptical_trenches import (
    grating_coupler_elliptical_trenches,
    grating_coupler_te,
    grating_coupler_tm,
)
from gdsfactory.components.grating_coupler_loss import (
    connect_loopback,
    grating_coupler_loss,
    loss_deembedding_ch12_34,
    loss_deembedding_ch13_24,
    loss_deembedding_ch14_23,
)
from gdsfactory.components.grating_coupler_rectangular import (
    grating_coupler_rectangular,
)
from gdsfactory.components.grating_coupler_rectangular_arbitrary import (
    grating_coupler_rectangular_arbitrary,
)
from gdsfactory.components.grating_coupler_rectangular_arbitrary_slab import (
    grating_coupler_rectangular_arbitrary_slab,
)
from gdsfactory.components.grating_coupler_tree import grating_coupler_tree
from gdsfactory.components.hline import hline
from gdsfactory.components.L import L
from gdsfactory.components.litho_calipers import litho_calipers
from gdsfactory.components.litho_ruler import litho_ruler
from gdsfactory.components.litho_steps import litho_steps
from gdsfactory.components.logo import logo
from gdsfactory.components.loop_mirror import loop_mirror
from gdsfactory.components.manhattan_font import manhattan_text
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.components.mmi2x2 import mmi2x2
from gdsfactory.components.mzi import mzi, mzi1x2_2x2, mzi2x2_2x2, mzi_coupler
from gdsfactory.components.mzi_arm import mzi_arm
from gdsfactory.components.mzi_arms import mzi_arms
from gdsfactory.components.mzi_lattice import mzi_lattice
from gdsfactory.components.mzi_phase_shifter import (
    mzi_phase_shifter,
    mzi_phase_shifter_top_heater_metal,
)
from gdsfactory.components.mzit import mzit
from gdsfactory.components.mzit_lattice import mzit_lattice
from gdsfactory.components.nxn import nxn
from gdsfactory.components.pad import pad, pad_array, pad_array90, pad_array270
from gdsfactory.components.pad_gsg import pad_gsg_open, pad_gsg_short
from gdsfactory.components.pads_shorted import pads_shorted
from gdsfactory.components.pcm_optical import pcm_optical
from gdsfactory.components.ramp import ramp
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.rectangle_with_slits import rectangle_with_slits
from gdsfactory.components.resistance_meander import resistance_meander
from gdsfactory.components.resistance_sheet import resistance_sheet
from gdsfactory.components.ring import ring
from gdsfactory.components.ring_double import ring_double
from gdsfactory.components.ring_single import ring_single
from gdsfactory.components.ring_single_array import ring_single_array
from gdsfactory.components.ring_single_dut import ring_single_dut, taper2
from gdsfactory.components.seal_ring import seal_ring
from gdsfactory.components.spiral import spiral
from gdsfactory.components.spiral_circular import spiral_circular
from gdsfactory.components.spiral_external_io import spiral_external_io
from gdsfactory.components.spiral_inner_io import (
    spiral_inner_io,
    spiral_inner_io_fiber_single,
)
from gdsfactory.components.splitter_chain import splitter_chain
from gdsfactory.components.splitter_tree import (
    splitter_tree,
    test_splitter_tree_ports,
    test_splitter_tree_ports_no_sbend,
)
from gdsfactory.components.straight import straight
from gdsfactory.components.straight_array import straight_array
from gdsfactory.components.straight_heater_doped_rib import straight_heater_doped_rib
from gdsfactory.components.straight_heater_doped_strip import (
    straight_heater_doped_strip,
)
from gdsfactory.components.straight_heater_meander import straight_heater_meander
from gdsfactory.components.straight_heater_metal import (
    straight_heater_metal,
    straight_heater_metal_90_90,
    straight_heater_metal_undercut,
    straight_heater_metal_undercut_90_90,
    test_ports,
)
from gdsfactory.components.straight_pin import straight_pin, straight_pn
from gdsfactory.components.straight_pin_slot import straight_pin_slot
from gdsfactory.components.straight_rib import straight_rib, straight_rib_tapered
from gdsfactory.components.switch_tree import switch_tree
from gdsfactory.components.taper import (
    taper,
    taper_strip_to_ridge,
    taper_strip_to_ridge_trenches,
)
from gdsfactory.components.taper_cross_section import (
    taper_cross_section_linear,
    taper_cross_section_sine,
)
from gdsfactory.components.taper_from_csv import (
    taper_0p5_to_3_l36,
    taper_from_csv,
    taper_w10_l100,
    taper_w10_l150,
    taper_w10_l200,
    taper_w11_l200,
    taper_w12_l200,
)
from gdsfactory.components.taper_parabolic import taper_parabolic
from gdsfactory.components.text import githash, text
from gdsfactory.components.text_rectangular import text_rectangular
from gdsfactory.components.triangle import triangle
from gdsfactory.components.verniers import verniers
from gdsfactory.components.version_stamp import pixel, qrcode, version_stamp
from gdsfactory.components.via import via, via1, via2, viac
from gdsfactory.components.via_cutback import via_cutback
from gdsfactory.components.waveguide_template import strip
from gdsfactory.components.wire import wire_corner, wire_straight
from gdsfactory.components.wire_sbend import wire_sbend

factory = dict(
    C=C,
    L=L,
    add_frame=add_frame,
    align_wafer=align_wafer,
    array=array,
    array_with_fanout=array_with_fanout,
    array_with_fanout_2d=array_with_fanout_2d,
    array_with_via=array_with_via,
    array_with_via_2d=array_with_via_2d,
    awg=awg,
    bbox=bbox,
    bend_circular=bend_circular,
    bend_circular180=bend_circular180,
    bend_circular_heater=bend_circular_heater,
    bend_euler=bend_euler,
    bend_euler180=bend_euler180,
    bend_euler_s=bend_euler_s,
    bend_straight_bend=bend_straight_bend,
    bend_port=bend_port,
    bend_s=bend_s,
    cavity=cavity,
    copy_layers=copy_layers,
    cdc=cdc,
    circle=circle,
    compass=compass,
    compensation_path=compensation_path,
    component_lattice=component_lattice,
    component_sequence=component_sequence,
    coupler=coupler,
    coupler90=coupler90,
    coupler90bend=coupler90bend,
    coupler90circular=coupler90circular,
    coupler_adiabatic=coupler_adiabatic,
    coupler_asymmetric=coupler_asymmetric,
    coupler_full=coupler_full,
    coupler_ring=coupler_ring,
    coupler_straight=coupler_straight,
    coupler_symmetric=coupler_symmetric,
    cross=cross,
    crossing=crossing,
    crossing45=crossing45,
    crossing_arm=crossing_arm,
    crossing_etched=crossing_etched,
    crossing_from_taper=crossing_from_taper,
    cutback_bend=cutback_bend,
    cutback_bend180=cutback_bend180,
    cutback_bend180circular=cutback_bend180circular,
    cutback_bend90=cutback_bend90,
    cutback_bend90circular=cutback_bend90circular,
    cutback_component=cutback_component,
    cutback_component_mirror=cutback_component_mirror,
    dicing_lane=dicing_lane,
    dbr=dbr,
    dbr_tapered=dbr_tapered,
    delay_snake=delay_snake,
    delay_snake2=delay_snake2,
    delay_snake3=delay_snake3,
    die=die,
    die_bbox=die_bbox,
    disk=disk,
    ellipse=ellipse,
    extend_port=extend_port,
    extend_ports=extend_ports,
    extend_ports_list=extend_ports_list,
    fiber=fiber,
    fiber_array=fiber_array,
    grating_coupler_array=grating_coupler_array,
    grating_coupler_elliptical=grating_coupler_elliptical,
    grating_coupler_circular=grating_coupler_circular,
    grating_coupler_circular_arbitrary=grating_coupler_circular_arbitrary,
    grating_coupler_elliptical_te=grating_coupler_elliptical_te,
    grating_coupler_elliptical_tm=grating_coupler_elliptical_tm,
    grating_coupler_elliptical_arbitrary=grating_coupler_elliptical_arbitrary,
    grating_coupler_elliptical_lumerical=grating_coupler_elliptical_lumerical,
    grating_coupler_elliptical_trenches=grating_coupler_elliptical_trenches,
    grating_coupler_loss=grating_coupler_loss,
    grating_coupler_te=grating_coupler_te,
    grating_coupler_tm=grating_coupler_tm,
    grating_coupler_tree=grating_coupler_tree,
    grating_coupler_rectangular=grating_coupler_rectangular,
    grating_coupler_rectangular_arbitrary=grating_coupler_rectangular_arbitrary,
    grating_coupler_rectangular_arbitrary_slab=grating_coupler_rectangular_arbitrary_slab,
    hline=hline,
    litho_calipers=litho_calipers,
    litho_steps=litho_steps,
    logo=logo,
    loop_mirror=loop_mirror,
    loss_deembedding_ch12_34=loss_deembedding_ch12_34,
    loss_deembedding_ch13_24=loss_deembedding_ch13_24,
    loss_deembedding_ch14_23=loss_deembedding_ch14_23,
    manhattan_text=manhattan_text,
    mmi1x2=mmi1x2,
    mmi2x2=mmi2x2,
    mzi=mzi,
    mzi2x2_2x2=mzi2x2_2x2,
    mzi1x2_2x2=mzi1x2_2x2,
    mzi_coupler=mzi_coupler,
    mzi_arm=mzi_arm,
    mzi_arms=mzi_arms,
    mzi_lattice=mzi_lattice,
    mzi_phase_shifter=mzi_phase_shifter,
    mzi_phase_shifter_top_heater_metal=mzi_phase_shifter_top_heater_metal,
    mzit=mzit,
    mzit_lattice=mzit_lattice,
    nxn=nxn,
    pad=pad,
    pad_gsg_short=pad_gsg_short,
    pad_gsg_open=pad_gsg_open,
    pad_array=pad_array,
    pads_shorted=pads_shorted,
    pcm_optical=pcm_optical,
    pixel=pixel,
    qrcode=qrcode,
    ramp=ramp,
    rectangle=rectangle,
    rectangle_with_slits=rectangle_with_slits,
    resistance_meander=resistance_meander,
    resistance_sheet=resistance_sheet,
    ring=ring,
    ring_double=ring_double,
    ring_single=ring_single,
    ring_single_array=ring_single_array,
    ring_single_dut=ring_single_dut,
    spiral=spiral,
    spiral_circular=spiral_circular,
    spiral_external_io=spiral_external_io,
    spiral_inner_io=spiral_inner_io,
    spiral_inner_io_fiber_single=spiral_inner_io_fiber_single,
    splitter_chain=splitter_chain,
    splitter_tree=splitter_tree,
    staircase=staircase,
    straight=straight,
    straight_array=straight_array,
    straight_heater_doped_rib=straight_heater_doped_rib,
    straight_heater_doped_strip=straight_heater_doped_strip,
    straight_heater_metal=straight_heater_metal,
    straight_heater_metal_90_90=straight_heater_metal_90_90,
    straight_heater_metal_undercut=straight_heater_metal_undercut,
    straight_heater_metal_undercut_90_90=straight_heater_metal_undercut_90_90,
    straight_heater_meander=straight_heater_meander,
    straight_pin=straight_pin,
    straight_pn=straight_pn,
    straight_pin_slot=straight_pin_slot,
    straight_rib=straight_rib,
    straight_rib_tapered=straight_rib_tapered,
    switch_tree=switch_tree,
    taper_cross_section_linear=taper_cross_section_linear,
    taper_cross_section_sine=taper_cross_section_sine,
    taper=taper,
    taper_parabolic=taper_parabolic,
    taper2=taper2,
    taper_0p5_to_3_l36=taper_0p5_to_3_l36,
    taper_from_csv=taper_from_csv,
    taper_strip_to_ridge=taper_strip_to_ridge,
    taper_strip_to_ridge_trenches=taper_strip_to_ridge_trenches,
    taper_w10_l100=taper_w10_l100,
    taper_w10_l150=taper_w10_l150,
    taper_w10_l200=taper_w10_l200,
    taper_w11_l200=taper_w11_l200,
    taper_w12_l200=taper_w12_l200,
    text=text,
    text_rectangular=text_rectangular,
    triangle=triangle,
    verniers=verniers,
    version_stamp=version_stamp,
    via=via,
    viac=viac,
    via1=via1,
    via2=via2,
    via_cutback=via_cutback,
    contact=contact,
    contact_slot=contact_slot,
    contact_slot_m1_m2=contact_slot_m1_m2,
    contact_heater_m3=contact_heater_m3,
    contact_slab_m3=contact_slab_m3,
    contact_with_offset=contact_with_offset,
    wire_corner=wire_corner,
    wire_sbend=wire_sbend,
    wire_straight=wire_straight,
    seal_ring=seal_ring,
)

_factory_passives = dict(
    bend_circular=bend_circular,
    bend_euler=bend_euler,
    bend_euler_s=bend_euler_s,
    bend_s=bend_s,
    cdc=cdc,
    coupler=coupler,
    coupler_adiabatic=coupler_adiabatic,
    coupler_asymmetric=coupler_asymmetric,
    coupler_full=coupler_full,
    coupler_ring=coupler_ring,
    coupler_symmetric=coupler_symmetric,
    crossing=crossing,
    crossing45=crossing45,
    taper_cross_section_linear=taper_cross_section_linear,
    taper_cross_section_sine=taper_cross_section_sine,
    taper=taper,
    taper2=taper2,
    taper_0p5_to_3_l36=taper_0p5_to_3_l36,
    taper_from_csv=taper_from_csv,
    taper_strip_to_ridge=taper_strip_to_ridge,
    taper_strip_to_ridge_trenches=taper_strip_to_ridge_trenches,
    taper_w10_l100=taper_w10_l100,
    taper_w10_l150=taper_w10_l150,
    taper_w10_l200=taper_w10_l200,
    taper_w11_l200=taper_w11_l200,
    taper_w12_l200=taper_w12_l200,
    mmi1x2=mmi1x2,
    mmi2x2=mmi2x2,
)
__all__ = [
    "factory",
    "C",
    "L",
    "add_frame",
    "align",
    "align_wafer",
    "array",
    "array_with_fanout",
    "array_with_fanout_2d",
    "array_with_via",
    "array_with_via_2d",
    "awg",
    "bbox",
    "bend_circular",
    "bend_circular180",
    "bend_circular_heater",
    "bend_euler",
    "bend_euler180",
    "bend_euler_s",
    "bend_port",
    "bend_s",
    "big_square",
    "cavity",
    "circle",
    "compass",
    "compensation_path",
    "component_lattice",
    "component_sequence",
    "connect_loopback",
    "coupler",
    "coupler90",
    "coupler90bend",
    "coupler90circular",
    "coupler_adiabatic",
    "coupler_asymmetric",
    "coupler_full",
    "coupler_ring",
    "coupler_straight",
    "coupler_symmetric",
    "cross",
    "crossing",
    "crossing45",
    "crossing_arm",
    "crossing_etched",
    "crossing_from_taper",
    "crossing_waveguide",
    "cutback_bend",
    "cutback_bend180",
    "cutback_bend180circular",
    "cutback_bend90",
    "cutback_bend90circular",
    "cutback_component",
    "cutback_component_mirror",
    "dbr",
    "dbr_tapered",
    "delay_snake",
    "delay_snake2",
    "delay_snake3",
    "die",
    "die_bbox",
    "disk",
    "ellipse",
    "ellipse_arc",
    "extend_port",
    "extend_ports",
    "extend_ports_list",
    "extension",
    "fiber",
    "fiber_array",
    "githash",
    "grating_coupler_array",
    "grating_coupler_elliptical",
    "grating_coupler_elliptical_arbitrary",
    "grating_coupler_elliptical_lumerical",
    "grating_coupler_circular",
    "grating_coupler_elliptical_te",
    "grating_coupler_elliptical_tm",
    "grating_coupler_elliptical_trenches",
    "grating_coupler_functions",
    "grating_coupler_loss",
    "grating_coupler_te",
    "grating_coupler_tm",
    "grating_coupler_tree",
    "grating_coupler_rectangular",
    "grating_coupler_rectangular_arbitrary",
    "grating_taper_points",
    "grating_tooth_points",
    "hline",
    "litho_calipers",
    "litho_ruler",
    "litho_steps",
    "logo",
    "loop_mirror",
    "loss_deembedding_ch12_34",
    "loss_deembedding_ch13_24",
    "loss_deembedding_ch14_23",
    "manhattan_font",
    "manhattan_text",
    "mmi1x2",
    "mmi2x2",
    "mzi",
    "mzi_arm",
    "mzi_lattice",
    "mzi_phase_shifter",
    "mzi_phase_shifter_top_heater_metal",
    "mzit",
    "mzit_lattice",
    "nxn",
    "pcm_optical",
    "pad",
    "pad_array",
    "pad_array90",
    "pad_array270",
    "pads_shorted",
    "pixel",
    "qrcode",
    "ramp",
    "rectangle",
    "resistance_meander",
    "ring",
    "ring_double",
    "ring_single",
    "ring_single_array",
    "ring_single_dut",
    "spiral",
    "spiral_circular",
    "spiral_external_io",
    "spiral_inner_io",
    "splitter_chain",
    "splitter_tree",
    "staircase",
    "straight",
    "straight_array",
    "straight_heater_doped_rib",
    "straight_heater_doped_strip",
    "straight_heater_metal",
    "straight_heater_metal_90_90",
    "straight_heater_metal_undercut",
    "straight_heater_metal_undercut_90_90",
    "straight_pin",
    "straight_pn",
    "straight_rib",
    "strip",
    "taper",
    "taper2",
    "taper_0p5_to_3_l36",
    "taper_from_csv",
    "taper_strip_to_ridge",
    "taper_strip_to_ridge_trenches",
    "taper_w10_l100",
    "taper_w10_l150",
    "taper_w10_l200",
    "taper_w11_l200",
    "taper_w12_l200",
    "test_delay_snake2_length",
    "test_delay_snake3_length",
    "test_ports",
    "test_splitter_tree_ports",
    "test_splitter_tree_ports_no_sbend",
    "text",
    "triangle",
    "verniers",
    "version_stamp",
    "via",
    "viac",
    "via1",
    "via2",
    "via_cutback",
    "contact",
    "contact_heater_m3",
    "contact_slab_m3",
    "contact_with_offset",
    "waveguide_template",
    "wire",
    "wire_corner",
    "wire_sbend",
    "wire_straight",
]



import tidy3d as td
import tidy3d.material_library as mat

MATERIALS = [m for m in dir(mat) if not m.startswith("_")]


def get_index(
    wavelength: float = 1.55,
    name: str = "cSi",
) -> float:
    """Returns refractive index from material database.

    Args: wavelength: wavelength (um) name: material name

    """
    if name not in MATERIALS:
        raise ValueError(f"{name} not in {MATERIALS}")

    medium = getattr(mat, name)
    medium.name = name
    frequency = 3e8 / wavelength
    return medium().epsilon(frequency)


def get_material(
    wavelength: float = 1.55,
    name: str = "cSi",
) -> td.Medium:
    """Returns Medium from materials database

    Args:
        wavelength: wavelength (um)
        name: material name

    """
    if name not in MATERIALS:
        raise ValueError(f"{name} not in {MATERIALS}")

    # FIXME: need to remove this.
    if name == "SiO2":
        return td.Medium(epsilon=1.45)
    elif name in ["cSi", "si"]:
        return td.Medium(n=3.48)
    elif name in ["SiN", "Si3N4"]:
        return td.Medium(n=2.0)
    else:
        raise ValueError(f"not implemetned material, name = {name}")

    medium = getattr(mat, name)
    medium.name = name
    return medium


if __name__ == "__main__":
    # print(get_index(name="aSi"))
    # SiO2 = get_material(name="SiO2")
    pass



""" CD SEM structures
"""
from functools import partial
from typing import List, Optional, Tuple

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bend_circular import bend_circular
from gdsfactory.components.manhattan_font import manhattan_text
from gdsfactory.components.straight import straight
from gdsfactory.cross_section import strip
from gdsfactory.tech import LAYER
from gdsfactory.types import ComponentFactory, CrossSectionFactory, Layer

LINE_LENGTH = 420.0


@cell
def cdsem_straight_all(
    straight: ComponentFactory = straight,
    cross_section: CrossSectionFactory = strip,
    layer: Tuple[int, int] = LAYER.WG,
    layers_cladding: List[Tuple[int, int]] = None,
    widths: Tuple[float, ...] = (0.4, 0.45, 0.5, 0.6, 0.8, 1.0),
    pixel_size: float = 1.0,
    length: float = LINE_LENGTH,
    spacing: float = 4.5,
) -> Component:
    c = Component()
    for width in widths:
        cross_section = partial(cross_section, width=width, layer=layer)
        c.add_ref(straight(length=length, cross_section=cross_section))

    c.distribute(direction="y", spacing=spacing)
    return c


@cell
def cdsem_straight_density(
    width: float = 0.372,
    trench_width: float = 0.304,
    x: float = LINE_LENGTH,
    y: float = 50.0,
    margin: float = 2.0,
    label: str = "",
    straight: ComponentFactory = straight,
    layer: Tuple[int, int] = LAYER.WG,
    layers_cladding: Optional[Tuple[Layer, ...]] = None,
    cross_section: CrossSectionFactory = strip,
    pixel_size: float = 1.0,
) -> Component:
    """Horizontal grating etch lines

    Args:
        width: width
        trench_width: trench_width
    """
    c = Component()
    period = width + trench_width
    n_o_lines = int((y - 2 * margin) / period)
    length = x - 2 * margin

    cross_section = partial(cross_section, width=width, layer=layer)
    tooth = straight(length=length, cross_section=cross_section)

    for i in range(n_o_lines):
        tooth_ref = c.add_ref(tooth)
        tooth_ref.movey((-n_o_lines / 2 + 0.5 + i) * period)
        c.absorb(tooth_ref)

    marker_label = manhattan_text(
        text=f"{label}",
        size=pixel_size,
        layer=layer,
    )
    _marker_label = c.add_ref(marker_label)
    _marker_label.move((length + 3, 10.0))
    c.absorb(_marker_label)
    return c


@cell
def cdsem_uturn(
    width: float = 0.5,
    radius: float = 10.0,
    wg_length: float = LINE_LENGTH,
    straight: ComponentFactory = straight,
    bend90: ComponentFactory = bend_circular,
    layer: Tuple[int, int] = LAYER.WG,
    layers_cladding: List[Tuple[int, int]] = None,
    cross_section: CrossSectionFactory = strip,
    pixel_size: float = 1.0,
) -> Component:
    """

    Args:
        width: of the line
        cladding_offset:
        radius: bend radius
        wg_length

    """
    c = Component()
    r = radius

    cross_section = partial(cross_section, width=width, layer=layer)
    if wg_length is None:
        wg_length = 2 * r

    bend90 = bend90(cross_section=cross_section, radius=r)
    wg = straight(
        cross_section=cross_section,
        length=wg_length,
    )

    # Add the U-turn on straight layer
    b1 = c.add_ref(bend90)
    b2 = c.add_ref(bend90)
    b2.connect("o2", b1.ports["o1"])

    wg1 = c.add_ref(wg)
    wg1.connect("o1", b1.ports["o2"])

    wg2 = c.add_ref(wg)
    wg2.connect("o1", b2.ports["o1"])

    label = c << manhattan_text(
        text=str(int(width * 1e3)), size=pixel_size, layer=layer
    )
    label.ymax = b2.ymin - 5
    label.x = 0
    b1.rotate(90)
    b2.rotate(90)
    wg1.rotate(90)
    wg2.rotate(90)
    label.rotate(90)
    return c


@cell
def pcm_optical(
    widths: Tuple[float, ...] = (0.4, 0.45, 0.5, 0.6, 0.8, 1.0),
    dense_lines_width: float = 0.3,
    dense_lines_width_difference: float = 20e-3,
    dense_lines_gap: float = 0.3,
    dense_lines_labels: Tuple[str, ...] = ("DL", "DM", "DH"),
    straight: ComponentFactory = straight,
    bend90: ComponentFactory = bend_circular,
    layer: Tuple[int, int] = LAYER.WG,
    layers_cladding: List[Tuple[int, int]] = None,
    cross_section: CrossSectionFactory = strip,
    pixel_size: float = 1.0,
) -> Component:
    """column with all optical PCMs

    Args:
        widths:  for straight

    """
    c = Component()
    _c1 = cdsem_straight_all(
        straight=straight,
        layer=layer,
        layers_cladding=layers_cladding,
        widths=widths,
        cross_section=cross_section,
        pixel_size=pixel_size,
    )

    all_devices = [_c1]

    all_devices += [
        cdsem_uturn(
            width=width,
            straight=straight,
            bend90=bend90,
            layer=layer,
            layers_cladding=layers_cladding,
            cross_section=cross_section,
            pixel_size=pixel_size,
        )
        for width in widths
    ]

    density_params = [
        (
            dense_lines_width - dense_lines_width_difference,
            dense_lines_gap - dense_lines_width_difference,
            dense_lines_labels[0],
        ),
        (dense_lines_width, dense_lines_gap, dense_lines_labels[1]),
        (
            dense_lines_width + dense_lines_width_difference,
            dense_lines_gap + dense_lines_width_difference,
            dense_lines_labels[2],
        ),
    ]

    all_devices += [
        cdsem_straight_density(
            width=w,
            trench_width=t,
            label=lbl,
            straight=straight,
            layer=layer,
            layers_cladding=layers_cladding,
            cross_section=cross_section,
            pixel_size=pixel_size,
        )
        for w, t, lbl in density_params
    ]

    [c.add_ref(d) for d in all_devices]
    c.align(elements="all", alignment="xmin")
    c.distribute(elements="all", direction="y", spacing=5, separation=True)
    return c


if __name__ == "__main__":
    # c = cdsem_straight()
    # c = cdsem_straight_all()
    # c = cdsem_uturn(width=2)
    # c = cdsem_straight_density()
    c = pcm_optical(layer=(2, 0))
    c.show()



from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.component import Component


def test_route_south(
    data_regression: DataRegressionFixture, check: bool = True
) -> Component:
    c = gf.Component("test_route_south")
    cr = c << gf.components.mmi2x2()
    route = gf.routing.route_south(component=cr)
    references = route.references

    lengths = {}
    for i, reference in enumerate(references):
        c.add(reference)
        route_length = reference.parent.info.length
        lengths[i] = route_length
    if check:
        data_regression.check(lengths)
    return c


if __name__ == "__main__":
    c = test_route_south(None, check=False)
    c.show()



from typing import Optional, Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.contact import contact_m1_m3 as contact_metal_function
from gdsfactory.components.contact import contact_slab_npp_m3
from gdsfactory.components.taper_cross_section import taper_cross_section
from gdsfactory.cross_section import rib_heater_doped, strip_rib_tip
from gdsfactory.snap import snap_to_grid
from gdsfactory.types import ComponentFactory, ComponentOrFactory, CrossSectionFactory


@gf.cell
def straight_heater_doped_rib(
    length: float = 320.0,
    nsections: int = 3,
    cross_section: CrossSectionFactory = strip_rib_tip,
    cross_section_heater: CrossSectionFactory = rib_heater_doped,
    contact: Optional[ComponentFactory] = contact_slab_npp_m3,
    contact_metal: Optional[ComponentFactory] = contact_metal_function,
    contact_metal_size: Tuple[float, float] = (10.0, 10.0),
    contact_size: Tuple[float, float] = (10.0, 10.0),
    taper: Optional[ComponentOrFactory] = taper_cross_section,
    heater_width: float = 2.0,
    heater_gap: float = 0.8,
    width: float = 0.5,
    with_top_contact: bool = True,
    with_bot_contact: bool = True,
    **kwargs
) -> Component:
    r"""Returns a doped thermal phase shifter.
    dimensions from https://doi.org/10.1364/OE.27.010456

    Args:
        length: of the waveguide
        nsections: between contacts
        cross_section: for the input/output ports
        cross_section_heater: for the heater
        contact: function to connect the heated strip
        contact_metal: function to connect the metal area
        contact_metal_size:
        contact_size:
        taper: optional taper function
        heater_width:
        heater_gap:
        width: waveguide width on the ridge
        kwargs: cross_section settings

    .. code::

                              length
        |<--------------------------------------------->|
        |              length_section                   |
        |    <--------------------------->              |
        |  length_contact                               |
        |    <------->                             taper|
        |    _________                    _________     |
        |   |        |                    |        |    |
        |   | contact|____________________|        |    |
        |   |  size  |    heater width    |        |    |
        |  /|________|____________________|________|\   |
        | / |             heater_gap               | \  |
        |/  |______________________________________|  \ |
         \  |_______________width__________________|  /
          \ |                                      | /
           \|_____________heater_gap______________ |/
            |        |                    |        |
            |        |____heater_width____|        |
            |        |                    |        |
            |________|                    |________|

        taper         cross_section_heater



                                   |<------width------>|
                                    ____________________ heater_gap             slab_gap
             top_contact           |                   |<---------->| bot_contact   <-->
         ___ ______________________|                   |___________________________|___
        |   |            |               undoped Si                 |              |   |
        |   |layer_heater|               intrinsic region           |layer_heater  |   |
        |___|____________|__________________________________________|______________|___|
                                                                     <------------>
                                                                      heater_width
        <------------------------------------------------------------------------------>
                                       slab_width

    """
    c = Component()
    cross_section_heater = gf.partial(
        cross_section_heater,
        heater_width=heater_width,
        heater_gap=heater_gap,
        width=width,
        **kwargs
    )

    if taper:
        taper = (
            taper(cross_section1=cross_section, cross_section2=cross_section_heater)
            if callable(taper)
            else taper
        )
        length -= taper.get_ports_xsize() * 2

    wg = c << gf.c.straight(
        cross_section=cross_section_heater,
        length=snap_to_grid(length),
    )

    if taper:
        taper1 = c << taper
        taper2 = c << taper
        taper2.mirror()
        taper1.connect("o2", wg.ports["o1"])
        taper2.connect("o2", wg.ports["o2"])
        c.add_port("o1", port=taper1.ports["o1"])
        c.add_port("o2", port=taper2.ports["o1"])

    else:
        c.add_port("o1", port=wg.ports["o1"])
        c.add_port("o2", port=wg.ports["o2"])

    if contact_metal:
        contact_section = contact_metal(size=contact_metal_size)
    contacts = []
    length_contact = snap_to_grid(contact_size[1])
    length_section = snap_to_grid((length - length_contact) / nsections)
    x0 = contact_size[0] / 2
    for i in range(0, nsections + 1):
        xi = x0 + length_section * i

        if contact_metal and contact:
            contact_center = c.add_ref(contact_section)
            contact_center.x = xi
            contact_ref = c << contact_section
            contact_ref.x = xi
            contact_ref.y = (
                +contact_metal_size[1] if i % 2 == 0 else -contact_metal_size[1]
            )
            contacts.append(contact_ref)

        if contact:
            if with_top_contact:
                contact_top = c << contact(size=contact_size)
                contact_top.x = xi
                contact_top.ymin = +(heater_gap + width / 2)

            if with_bot_contact:
                contact_bot = c << contact(size=contact_size)
                contact_bot.x = xi
                contact_bot.ymax = -(heater_gap + width / 2)

    if contact_metal and contact:
        contact_length = length + contact_metal_size[0]
        contact_top = c << contact_metal(
            size=(contact_length, contact_metal_size[0]),
        )
        contact_bot = c << contact_metal(
            size=(contact_length, contact_metal_size[0]),
        )

        contact_bot.xmin = contacts[0].xmin
        contact_top.xmin = contacts[0].xmin

        contact_top.ymin = contacts[0].ymax
        contact_bot.ymax = contacts[1].ymin

        c.add_ports(contact_top.ports, prefix="top_")
        c.add_ports(contact_bot.ports, prefix="bot_")
    return c


def test_straight_heater_doped_rib_ports() -> Component:
    c = straight_heater_doped_rib(length=100.0)
    assert c.get_ports_xsize(port_type="optical") == 100.0, c.get_ports_xsize(
        port_type="optical"
    )
    return c


if __name__ == "__main__":
    c = straight_heater_doped_rib(with_top_heater=False, with_top_contact=False)
    # c = straight_heater_doped_rib()
    c.show()



"""This is a sample on how to define custom components."""
import shutil
from pathlib import Path

import gdsfactory as gf
from gdsfactory.add_grating_couplers import (
    add_grating_couplers_with_loopback_fiber_array,
)
from gdsfactory.autoplacer.yaml_placer import place_from_yaml
from gdsfactory.component import Component
from gdsfactory.components.spiral_inner_io import spiral_inner_io
from gdsfactory.config import CONFIG, logger
from gdsfactory.sweep.write_sweeps import write_sweeps

add_te = gf.partial(
    gf.routing.add_fiber_array,
    grating_coupler=gf.components.grating_coupler_elliptical_te,
)


add_tm = gf.partial(
    gf.routing.add_fiber_array,
    grating_coupler=gf.components.grating_coupler_elliptical_tm,
)


@gf.cell
def coupler_te(gap: float, length: float) -> Component:
    """Directional coupler with TE grating couplers."""
    c = gf.components.coupler(gap=gap, length=length)
    cc = add_te(c)
    return cc


@gf.cell
def spiral_te(width: float = 0.5, length: float = 20e3) -> Component:
    """Waveguide Spiral with TE grating_coupler

    Args:
        width: um
        length: um
    """
    c = spiral_inner_io(width=width, length=length)
    c = gf.c.extend_ports(c)
    cc = add_grating_couplers_with_loopback_fiber_array(
        component=c,
        grating_coupler=gf.components.grating_coupler_elliptical_te,
        bend=gf.components.bend_circular,
    )
    return cc


@gf.cell
def spiral_tm(width: float = 0.5, length: float = 20e3) -> Component:
    """Waveguide Spiral with TM grating_coupler.

    Args:
        width: um
        lenght: um
    """
    c = spiral_inner_io(width=width, length=length, dx=10, dy=10, N=5)
    c = gf.c.extend_ports(c)
    cc = add_grating_couplers_with_loopback_fiber_array(
        component=c,
        grating_coupler=gf.components.grating_coupler_elliptical_tm,
        bend=gf.components.bend_circular,
    )
    return cc


component_factory = dict(
    spiral_te=spiral_te, spiral_tm=spiral_tm, coupler_te=coupler_te
)


def test_mask(precision: float = 2e-9) -> Path:
    workspace_folder = CONFIG["samples_path"] / "mask_custom"
    build_path = workspace_folder / "build"
    doe_root_path = build_path / "cache_doe"
    doe_metadata_path = build_path / "sweep"
    mask_path = build_path / "mask"
    does_yml = workspace_folder / "does.yml"

    shutil.rmtree(build_path, ignore_errors=True)
    mask_path.mkdir(parents=True, exist_ok=True)

    gdspath = mask_path / "sample_mask.gds"
    logpath = gdspath.with_suffix(".log")
    logger.add(sink=logpath)

    write_sweeps(
        str(does_yml),
        component_factory=component_factory,
        precision=precision,
        doe_root_path=doe_root_path,
        doe_metadata_path=doe_metadata_path,
    )

    top_level = place_from_yaml(does_yml, precision=precision, root_does=doe_root_path)
    top_level.write(str(gdspath))

    assert gdspath.exists()
    return gdspath


if __name__ == "__main__":
    # gdspath_mask = test_mask()
    # gf.show(gdspath_mask)
    # c = coupler_te(gap=0.3, length=2.0)
    # c = spiral_te(length=60e3)
    # c.show()

    # lengths = [18.24, 36.48, 54.72, 72.96, 91.2]
    # for length in lengths:
    #     c = coupler_te(gap=0.3, length=length)

    gds = test_mask()
    gf.show(gds)



from gdsfactory.simulation.gmeep.add_monitors import add_monitors
from gdsfactory.simulation.gmeep.get_simulation import get_simulation
from gdsfactory.simulation.gmeep.get_transmission_2ports import (
    get_transmission_2ports,
    plot2D,
    plot3D,
)
from gdsfactory.simulation.gmeep.plot_xsection import plot_xsection

__all__ = [
    "add_monitors",
    "get_simulation",
    "get_sparameters1x2",
    "get_transmission_2ports",
    "plot2D",
    "plot3D",
    "plot_xsection",
]
__version__ = "0.0.2"



import phidl.geometry as pg

import gdsfactory as gf
from gdsfactory.component import Component


@gf.cell
def outline(elements, **kwargs) -> Component:
    """
    Returns Component containing the outlined polygon(s).

    wraps phidl.geometry.outline

    Creates an outline around all the polygons passed in the `elements`
    argument. `elements` may be a Device, Polygon, or list of Devices.

    Args:
        elements: Device(/Reference), list of Device(/Reference), or Polygon
            Polygons to outline or Device containing polygons to outline.
        distance : int or float
            Distance to offset polygons. Positive values expand, negative shrink.
        precision : float
            Desired precision for rounding vertex coordinates.
        num_divisions : array-like[2] of int
            The number of divisions with which the geometry is divided into
            multiple rectangular regions. This allows for each region to be
            processed sequentially, which is more computationally efficient.
        join : {'miter', 'bevel', 'round'}
            Type of join used to create the offset polygon.
        tolerance : int or float
            For miter joints, this number must be at least 2 and it represents the
            maximal distance in multiples of offset between new vertices and their
            original position before beveling to avoid spikes at acute joints. For
            round joints, it indicates the curvature resolution in number of
            points per full circle.
        join_first : bool
            Join all paths before offsetting to avoid unnecessary joins in
            adjacent polygon sides.
        max_points : int
            The maximum number of vertices within the resulting polygon.
        open_ports : bool or float
            If not False, holes will be cut in the outline such that the Ports are
            not covered. If True, the holes will have the same width as the Ports.
            If a float, the holes will be be widened by that value (useful for fully
            clearing the outline around the Ports for positive-tone processes
        layer : int, array-like[2], or set
            Specific layer(s) to put polygon geometry on.)

    """
    return gf.read.from_phidl(component=pg.outline(elements, **kwargs))


def test_outline():
    e1 = gf.components.ellipse(radii=(6, 6))
    e2 = gf.components.ellipse(radii=(10, 4))
    c = outline([e1, e2])
    assert int(c.area()) == 52


if __name__ == "__main__":
    e1 = gf.components.ellipse(radii=(6, 6))
    e2 = gf.components.ellipse(radii=(10, 4))
    c = outline([e1, e2])
    c.show()



import meep as mp
import numpy as np
import pandas as pd

import gdsfactory as gf
from gdsfactory.simulation.gmeep.get_simulation import get_simulation

if __name__ == "__main__":
    c = gf.components.straight(length=2)
    sim_dict = get_simulation(component=c)
    sim = sim_dict["sim"]
    monitors = sim_dict["monitors"]
    freqs = sim_dict["freqs"]
    field_monitor_point = sim_dict["field_monitor_point"]
    wavelengths = 1 / freqs
    filepath = "sp.csv"
    sources = sim_dict["sources"]
    source = sources[0]
    source_power = source.eig_power(1 / 1.55)
    port_source_name = sim_dict["port_source_name"]

    # sim.plot2D()
    # plt.show()

    sim.run(
        until_after_sources=mp.stop_when_fields_decayed(
            dt=50, c=mp.Ez, pt=field_monitor_point, decay_by=1e-9
        )
    )

    # call this function every 50 time spes
    # look at simulation and measure component that we want to measure (Ez component)
    # when field_monitor_point decays below a certain 1e-9 field threshold

    # Calculate mode overlaps
    nports = len(monitors)
    r = dict(wavelengths=wavelengths)
    S = np.zeros((len(freqs), nports, nports))
    a = {}
    b = {}

    # Parse out the overlaps
    for port_name, monitor in monitors.items():
        m_results = np.abs(sim.get_eigenmode_coefficients(monitor, [1]).alpha)
        a[port_name] = m_results[:, :, 0]  # forward wave
        b[port_name] = m_results[:, :, 1]  # backward wave

    source_fields = np.squeeze(a[port_source_name])
    for i, port_name_i in enumerate(monitors.keys()):
        monitor = monitors[port_name_i]
        r[f"s{i+1}"] = sim.get_eigenmode_coefficients(monitor, [1]).alpha[0, :, 0]

        # for j, port_name_j in enumerate(monitors.keys()):
        #     if port_name_j == port_name_i:
        #         S[:, i, j] = np.squeeze(b[port_name_j]) / source_fields
        #     else:
        #         S[:, i, j] = np.squeeze(a[port_name_j]) / source_fields

        # S[:, i, j] = np.squeeze(a[port_name_j] / b[port_name_i])
        # S[:, j, i] = np.squeeze(a[port_name_i] / b[port_name_j])

    # for port_name in monitor.keys():
    #     a1 = m1_results[:, :, 0]  # forward wave
    #     b1 = m1_results[:, :, 1]  # backward wave
    #     a2 = m2_results[:, :, 0]  # forward wave
    #     # b2 = m2_results[:, :, 1]  # backward wave

    #     # Calculate the actual scattering parameters from the overlaps
    #     s11 = np.squeeze(b1 / a1)
    #     s12 = np.squeeze(a2 / a1)

    keys = [key for key in r.keys() if key.startswith("s")]
    s = {f"{key}a": list(np.unwrap(np.angle(r[key].flatten()))) for key in keys}
    s.update({f"{key}m": list(np.abs(r[key].flatten())) for key in keys})
    s.update(wavelengths=wavelengths)
    s.update(freqs=freqs)
    df = pd.DataFrame(s)

    # df = df.set_index(df.wavelength)
    # df.to_csv(filepath, index=False)
    # s11 = S[:, 0, 0]
    # s12 = S[:, 0, 1]
    # s11 = np.abs(s11)
    # s12 = np.abs(s12)
    # s11 = 10 * np.log10(s11)
    # s12 = 10 * np.log10(s12)
    # plt.plot(wavelengths, s11, label="s11")
    # plt.plot(wavelengths, s12, label="s12")
    # plt.legend()

    monitor_out = monitors[2]
    source_power = np.abs(source_fields) ** 2
    transmission = (
        np.abs(sim.get_eigenmode_coefficients(monitor_out, [1]).alpha[0, :, 0]) ** 2
        / source_power
    )
    print(f"transmission: {transmission}")

    monitor_input = monitors[1]
    reflection = (
        np.abs(sim.get_eigenmode_coefficients(monitor_input, [1]).alpha[0, :, 1]) ** 2
        / source_power
    )
    print(f"Reflection: {reflection}")

    # sim.plot2D()
    # plt.show()
    # plt.show()



from typing import Optional

import numpy as np
import phidl.routing as pr

import gdsfactory as gf
from gdsfactory.port import Port


@gf.cell
def route_quad(
    port1: Port,
    port2: Port,
    width1: Optional[float] = None,
    width2: Optional[float] = None,
    layer: gf.types.Layer = (31, 0),
) -> gf.Component:
    """Routes a basic quadrilateral polygon directly between two ports.

    based on phidl.routing.route_quad

    Args:
        port1: Port to start route
        port2 : Port objects to end route
        width1: Width of quadrilateral at ports. If None, uses port widths.
        width2: Width of quadrilateral at ports. If None, uses port widths.
        layer: Layer to put the route on.

    """

    def get_port_edges(port, width):
        _, e1 = pr._get_rotated_basis(port.orientation)
        pt1 = port.midpoint + e1 * width / 2
        pt2 = port.midpoint - e1 * width / 2
        return pt1, pt2

    if width1 is None:
        width1 = port1.width
    if width2 is None:
        width2 = port2.width
    vertices = np.array(get_port_edges(port1, width1) + get_port_edges(port2, width2))
    center = np.mean(vertices, axis=0)
    displacements = vertices - center
    # sort vertices by angle from center of quadrilateral to make convex polygon
    angles = np.array([np.arctan2(disp[0], disp[1]) for disp in displacements])
    vertices = [vert for _, vert in sorted(zip(angles, vertices), key=lambda x: x[0])]
    component = gf.Component()
    component.add_polygon(points=vertices, layer=layer)
    component.add_port(
        name="e1",
        midpoint=port1.midpoint,
        orientation=port1.orientation + 180,
        width=width1,
    )
    component.add_port(
        name="e2",
        midpoint=port2.midpoint,
        orientation=port2.orientation + 180,
        width=width2,
    )
    return component


if __name__ == "__main__":
    c = gf.Component()
    mzi = c << gf.c.mzi_phase_shifter()
    pads = c << gf.c.array(component=gf.c.pad, columns=3)
    pads.ymin = mzi.ymax + 30

    pads.movex(-pads.size_info.sc[0])
    mzi.movex(-mzi.size_info.sc[0])

    route_gnd = c << route_quad(
        mzi.ports["e1"],
        pads.ports["e4_1_1"],
        width1=None,
        width2=None,
    )

    c.show()
